@@ -463,12 +463,16 @@ NOTE: This interface is experimental and subject to change."
 (define-load-time-global *cache-vector-symbols* nil)
 
 (defun drop-all-hash-caches ()
+  #+sb-xc-host (values-specifier-type-cache-clear) ; it's not like the rest
   (dolist (name *cache-vector-symbols*)
     (set name nil)))
 
+(defmacro sb-int-package () (find-package "SB-INT"))
+
 ;; Make a new hash-cache and optionally create the statistics vector.
 (defun alloc-hash-cache (size symbol)
   (declare (type index size))
+  (declare (sb-c::tlab :system))
   (let (cache)
     ;; It took me a while to figure out why infinite recursion could occur
     ;; in VALUES-SPECIFIER-TYPE. It's because SET calls VALUES-SPECIFIER-TYPE.
@@ -483,8 +487,7 @@ NOTE: This interface is experimental and subject to change."
                  ;; it is inconsequential to performance.
                  (if *profile-hash-cache*
                      `(let ((statistics
-                             (let ((*package* (sb-xc:symbol-package symbol)))
-                               (symbolicate symbol "STATISTICS"))))
+                             (package-symbolicate (sb-int-package) symbol "-STATS")))
                         (unless (boundp statistics)
                           (set! statistics
                                 (make-array 3 :element-type 'fixnum
@@ -520,6 +523,7 @@ NOTE: This interface is experimental and subject to change."
                `(progn
                   (declaim (ftype ,ftype ,fn))
                   (defun ,fn ,args
+                    (declare (sb-c::tlab :system))
                     (declare (optimize (safety 0)))
                     ,(if (<= n 3)
                          `(list* ,@args)
@@ -530,50 +534,81 @@ NOTE: This interface is experimental and subject to change."
   (def 5)
   (def 6))
 
-(defmacro !define-hash-cache (name args aux-vars
+(defmacro !define-hash-cache (name args
                               &key hash-function hash-bits memoizer
-                              flush-function (values 1))
+                              (values 1))
   (declare (ignore memoizer))
   (dolist (arg args)
+    ;; Each arg is (VARNAME EQUIVALENCE-TEST &OPTIONAL EXPR-TO-CACHE)
+    ;; EXPR-TO-CACHE says how to copy a dynamic-extent arg into the cache line
+    ;; (the default is to store each argument as-is)
     (unless (<= 2 (length arg) 3)
       (error "bad argument spec: ~S" arg)))
-  (assert (typep hash-bits '(integer 5 14))) ; reasonable bounds
-  (let* ((fun-name (symbolicate "!" name "-MEMO-WRAPPER"))
-         (var-name (symbolicate "**" name "-CACHE-VECTOR**"))
+  ;; Disallow >12 hash bits. Supposing a 32-bit machine, a ctype hash uses 5
+  ;; reserved bits for the type-class index, and then we need 2 independent
+  ;; choices of hash, so supposing that's 12 bits each, we're at 29 bits.
+  ;; But I want to take 2 bits for flags, so that's 31 bits.
+  (assert (typep hash-bits '(integer 5 12))) ; reasonable bounds
+  (unless (integerp values) ; non-numeric was only for CTYPE-OF-ARRAY
+    (error "Values ~S unsupported" values))
+  (let* ((var-name (symbolicate "**" name "-CACHE-VECTOR**"))
+         (statistics-name (package-symbolicate (sb-int-package) var-name "-STATS"))
+         (cache-type `(simple-vector ,(ash 1 hash-bits))))
+    `(progn
+       (eval-when (:compile-toplevel :execute)
+         (setf (get ',name '!cache-definition)
+               (list ',hash-bits ',hash-function ',args ',values)))
+       (pushnew ',var-name *cache-vector-symbols*)
+       (define-load-time-global ,var-name nil)
+       ,@(when *profile-hash-cache*
+           `((declaim (type (simple-array fixnum (3)) ,statistics-name))
+             (defvar ,statistics-name)))
+       (declaim (type (or null ,cache-type) ,var-name))
+       (defun ,(symbolicate name "-CACHE-CLEAR") () (setq ,var-name nil))
+       ',var-name)))
+
+(eval-when (:compile-toplevel :execute) ; leave this out of the core image
+(#+sb-xc-host defmacro #-sb-xc-host sb-xc:defmacro
+ with-cache ((fun-name &rest actual-args) &body computation)
+  (let* ((var-name (package-symbolicate (cl:symbol-package fun-name)
+                                        "**" fun-name "-CACHE-VECTOR**"))
+         (cache-params (get fun-name '!cache-definition))
          (statistics-name
           (when *profile-hash-cache*
-            (symbolicate var-name "STATISTICS")))
-         (nargs (length args))
+            (package-symbolicate (sb-int-package) var-name "-STATS")))
+         (hash-bits (first cache-params))
+         (hashfun (second cache-params))
+         (arg-specs (third cache-params))
+         (nargs (length arg-specs))
+         (values (fourth cache-params))
          (size (ash 1 hash-bits))
          (hashval (make-symbol "HASH"))
          (cache (make-symbol "CACHE"))
          (entry (make-symbol "LINE"))
-         (thunk (make-symbol "THUNK"))
-         (arg-vars (mapcar #'first args))
          (nvalues (if (listp values) (length values) values))
          (result-temps
           (if (listp values)
               values ; use the names provided by the user
               (loop for i from 1 to nvalues ; else invent some names
                     collect (make-symbol (format nil "R~D" i)))))
-         (temps (append (mapcar (lambda (x) (make-symbol (string x)))
-                                arg-vars)
-                        result-temps))
+         (temps (append (mapcar #'copy-symbol actual-args) result-temps))
          ;; Mnemonic: (FIND x SEQ :test #'f) calls f with x as the LHS
-         (tests (mapcar (lambda (spec temp) ; -> (EQx ARG #:ARG)
-                          `(,(cadr spec) ,(car spec) ,temp))
-                        args temps))
+         (tests (mapcar (lambda (spec temp actual) ; -> (EQx #:ARG ARG)
+                          `(,(cadr spec) ,temp ,actual))
+                        arg-specs temps actual-args))
          (cache-type `(simple-vector ,size))
          (line-type (let ((n (+ nargs nvalues)))
                       (if (<= n 3) 'cons `(simple-vector ,n))))
-         (bind-hashval
-          `((,hashval (the sb-xc:fixnum
-                           (funcall ,hash-function ,@arg-vars)))
-            (,cache ,var-name)))
+         ;; It's not really necessary to factor out PROBE-IT,
+         ;; but it used to be possible to probe for an entry to forcefully
+         ;; evict (unmemoize) it. That capability is no longer needed.
          (probe-it
-          (lambda (ignore action)
+          (lambda (action)
             `(when ,cache
-               (let ((,hashval ,hashval) ; gets clobbered in probe loop
+               ;; Shift out 2 low bits because they won't be as random as other bits,
+               ;; at least not for single-argument cached functions. They might be OK
+               ;; for functions that involve mixing.
+               (let ((,hashval (ash ,hashval -2)) ; gets clobbered in probe loop
                      (,cache (truly-the ,cache-type ,cache)))
                  ;; FIXME: redundant?
                  (declare (type sb-xc:fixnum ,hashval))
@@ -582,78 +617,61 @@ NOTE: This interface is experimental and subject to change."
                               (svref ,cache
                                      (ldb (byte ,hash-bits 0) ,hashval))))
                          (unless (eql ,entry 0)
-                           ;; This barrier is a no-op on all multi-threaded SBCL
-                           ;; architectures. No CPU except Alpha will move a
-                           ;; load prior to a load on which it depends.
-                           (sb-thread:barrier (:data-dependency))
-                           (locally (declare (type ,line-type ,entry))
-                             (let* ,(case (length temps)
+                           (let* ((,entry (truly-the ,line-type ,entry))
+                                  ,@(case (length temps)
                                      (2 `((,(first temps) (car ,entry))
                                           (,(second temps) (cdr ,entry))))
-                                     (3 (let ((arg-temp (sb-xc:gensym "ARGS")))
+                                     (3 (let ((arg-temp (gensym "ARGS")))
                                           `((,arg-temp (cdr ,entry))
                                             (,(first temps) (car ,entry))
                                             (,(second temps)
                                              (car (truly-the cons ,arg-temp)))
                                             (,(third temps) (cdr ,arg-temp)))))
                                      (t (loop for i from 0 for x in temps
-                                              collect `(,x (svref ,entry ,i)))))
-                               ,@ignore
-                               (when (and ,@tests) ,action))))
-                         (setq ,hashval (ash ,hashval ,(- hash-bits)))))))))
-         (fun
-          `(defun ,fun-name (,thunk ,@arg-vars ,@aux-vars)
-             ,@(when *profile-hash-cache* ; count seeks
-                 `((when (boundp ',statistics-name)
-                     (incf (aref ,statistics-name 0)))))
-             (let ,bind-hashval
-               ,(funcall probe-it nil
-                         `(return-from ,fun-name (values ,@result-temps)))
-               (multiple-value-bind ,result-temps (funcall ,thunk)
-                 (let ((,entry
-                        (,(hash-cache-line-allocator (+ nargs nvalues))
-                         ,@(mapcar (lambda (spec) (or (caddr spec) (car spec)))
-                                   args)
-                         ,@result-temps))
-                       (,cache
-                        (truly-the ,cache-type
-                         (or ,cache (alloc-hash-cache ,size ',var-name))))
-                       (idx1 (ldb (byte ,hash-bits 0) ,hashval))
-                       (idx2 (ldb (byte ,hash-bits ,hash-bits) ,hashval)))
-                   ,@(when *profile-hash-cache*
-                       `((incf (aref ,statistics-name 1)))) ; count misses
+                                              collect `(,x (svref ,entry ,i))))))
+                               (when (and ,@tests) ,action)))
+                         (setq ,hashval (ash ,hashval ,(- hash-bits))))))))))
+    `(let ((,hashval (the sb-xc:fixnum (funcall ,hashfun ,@actual-args)))
+           (,cache ,var-name))
+       #-sb-xc-host (declare (optimize (sb-c::insert-array-bounds-checks 0)))
+       ,@(when *profile-hash-cache* ; count seeks
+           `((when (boundp ',statistics-name) (incf (aref ,statistics-name 0)))))
+       (block seek
+         ,(funcall probe-it `(return-from seek (values ,@result-temps)))
+         (multiple-value-bind ,result-temps (progn ,@computation)
+           ;; Decide if cacheable result. Lambda expressions and strings always are.
+           ;; Parsing a type specifier nonlocally exits from COMPUTATION if it
+           ;; wants not to cache. And of course CTYPE-OF always returns a good type.
+           ;; It's unfortunate that this macro macro bakes in such knowledge.
+           (when ,(if (member fun-name '(sb-format::tokenize-control-string
+                                         sb-alien::coerce-to-interpreted-function
+                                         values-specifier-type
+                                         ctype-of))
+                      t
+                      `(and ,@(mapcar (lambda (x) `(sb-kernel::ok-to-memoize-p ,x))
+                                      actual-args)))
+             (let ((,entry (,(hash-cache-line-allocator (+ nargs nvalues))
+                            ,@(mapcar (lambda (spec actual) (or (caddr spec) actual))
+                                      arg-specs actual-args)
+                            ,@result-temps))
+                   (,cache (truly-the ,cache-type
+                                      (or ,cache (alloc-hash-cache ,size ',var-name))))
+                   (idx1 (ldb (byte ,hash-bits 2) ,hashval))
+                   (idx2 (ldb (byte ,hash-bits ,(+ 2 hash-bits)) ,hashval)))
+               ,@(when *profile-hash-cache*
+                   `((incf (aref ,statistics-name 1)))) ; count misses
                    ;; Why a barrier: the pointer to 'entry' (a cons or vector)
                    ;; MUST NOT be observed by another thread before its cells
                    ;; are filled. Equally bad, the 'output' cells in the line
                    ;; could be 0 while the 'input' cells matched something.
-                   (sb-thread:barrier (:write))
-                   (cond ((eql (svref ,cache idx1) 0)
-                          (setf (svref ,cache idx1) ,entry))
-                         ((eql (svref ,cache idx2) 0)
-                          (setf (svref ,cache idx2) ,entry))
-                         (t
-                           ,@(when *profile-hash-cache* ; count evictions
-                               `((incf (aref ,statistics-name 2))))
-                           (setf (svref ,cache idx1) ,entry))))
-                 (values ,@result-temps))))))
-    `(progn
-       (pushnew ',var-name *cache-vector-symbols*)
-       (define-load-time-global ,var-name nil)
-       ,@(when *profile-hash-cache*
-           `((declaim (type (simple-array fixnum (3)) ,statistics-name))
-             (defvar ,statistics-name)))
-       (declaim (type (or null ,cache-type) ,var-name))
-       (defun ,(symbolicate name "-CACHE-CLEAR") () (setq ,var-name nil))
-       ,@(when flush-function
-           `((defun ,flush-function ,arg-vars
-               (let ,bind-hashval
-                 ,(funcall probe-it
-                   `((declare (ignore ,@result-temps)))
-                   `(return (setf (svref ,cache
-                                         (ldb (byte ,hash-bits 0) ,hashval))
-                                  0)))))))
-       (declaim (inline ,fun-name))
-       ,fun)))
+               (sb-thread:barrier (:write))
+               (setf (svref ,cache (cond ((eql (svref ,cache idx1) 0) idx1)
+                                         ((eql (svref ,cache idx2) 0) idx2)
+                                         (t ,@(when *profile-hash-cache* ; count evictions
+                                                `((incf (aref ,statistics-name 2))))
+                                            idx1)))
+                     ,entry)))
+           (values ,@result-temps)))))))
 
 ;;; some syntactic sugar for defining a function whose values are
 ;;; cached by !DEFINE-HASH-CACHE
@@ -680,10 +698,7 @@ NOTE: This interface is experimental and subject to change."
 ;;; Since we don't have caches that aren't in direct support of DEFUN-CACHED
 ;;; - did we ever? - this should be possible to change.
 ;;;
-(defmacro defun-cached ((name &rest options &key
-                              (memoizer (make-symbol "MEMOIZE")
-                                        memoizer-supplied-p)
-                              &allow-other-keys)
+(defmacro defun-cached ((name &rest options &key memoizer &allow-other-keys)
                         args &body body-decls-doc)
   (binding* (((forms decls doc) (parse-body body-decls-doc t))
              ((inputs aux-vars)
@@ -692,42 +707,30 @@ NOTE: This interface is experimental and subject to change."
                     (values (ldiff args aux) aux)
                     (values args nil))))
              (arg-names (mapcar #'car inputs)))
+    (assert (not aux-vars)) ; was only for CTYPE-OF-ARRAY
     `(progn
-        (!define-hash-cache ,name ,inputs ,aux-vars ,@options)
+        (!define-hash-cache ,name ,inputs ,@options)
         (defun ,name ,arg-names
           ,@decls
           ,@(if doc (list doc))
-          (macrolet ((,memoizer (&body body)
-                       ;; We don't need (DX-FLET ((,thunk () ,@body)) ...)
-                       ;; This lambda is a single-use local call within
-                       ;; the inline memoizing wrapper.
-                       `(,',(symbolicate "!" name "-MEMO-WRAPPER")
-                         (lambda () ,@body) ,@',arg-names)))
-             ,@(if memoizer-supplied-p
-                   forms
-                   `((,memoizer ,@forms))))))))
+          ,(if memoizer
+               `(macrolet ((,memoizer (&body body)
+                             `(with-cache ,',(list* name arg-names) ,@body)))
+                  ,@forms)
+               `(with-cache (,name ,@arg-names) ,@forms))))))
 
-;;; FIXME: maybe not the best place
-;;;
-;;; FIXME: think of a better name -- not only does this not have the
-;;; CAR recursion of EQUAL, it also doesn't have the special treatment
-;;; of pathnames, bit-vectors and strings.
-;;;
-;;; KLUDGE: This means that we will no longer cache specifiers of the
-;;; form '(INTEGER (0) 4).  This is probably not a disaster.
-;;;
-;;; A helper function for the type system, which is the main user of
-;;; these caches: we must be more conservative than EQUAL for some of
-;;; our equality tests, because MEMBER and friends refer to EQLity.
-;;; So:
-(defun equal-but-no-car-recursion (x y)
-  (do () (())
-    (cond ((eql x y) (return t))
-          ((and (consp x)
-                (consp y)
-                (eql (pop x) (pop y))))
-          (t
-           (return)))))
+(defun list-elts-eq (x y)
+  (loop (when (eq x y) (return t))
+        (when (or (atom x) (atom y)) (return nil))
+        (unless (eq (car x) (car y)) (return nil))
+        (setq x (cdr x)
+              y (cdr y))))
+(defun list-elements-eql (x y)
+  (loop (when (eq x y) (return t))
+        (when (or (atom x) (atom y)) (return nil))
+        (unless (eql (car x) (car y)) (return nil))
+        (setq x (cdr x)
+              y (cdr y))))
 
 ;;;; various operations on names
 
@@ -827,7 +830,7 @@ NOTE: This interface is experimental and subject to change."
           (let* ((name (first spec))
                  (exp-temp (gensym "ONCE-ONLY")))
             `(let ((,exp-temp ,(second spec))
-                   (,name (sb-xc:gensym ,(symbol-name name))))
+                   (,name (gensym ,(symbol-name name))))
                `(let ((,,name ,,exp-temp))
                   ,,(frob (rest specs) body))))))))
 
@@ -1344,8 +1347,7 @@ NOTE: This interface is experimental and subject to change."
 '(defun show-hash-cache-statistics ()
   (flet ((cache-stats (symbol)
            (let* ((name (string symbol))
-                  (statistics (let ((*package* (sb-xc:symbol-package symbol)))
-                                (symbolicate symbol "STATISTICS")))
+                  (statistics (package-symbolicate (sb-int-package) symbol "-STATS"))
                   (prefix
                    (subseq name 0 (- (length name) (length "VECTOR**")))))
              (values (if (boundp statistics)
@@ -1374,6 +1376,16 @@ NOTE: This interface is experimental and subject to change."
                                 (length cache))))
                   short-name))))))
 
+;;; some commonly-occurring CONSTANTLY forms
+(macrolet ((def-constantly-fun (name constant-expr)
+             `(defun ,name (&rest arguments)
+                (declare (ignore arguments))
+                (declare (optimize (speed 3) (safety 0) (debug 0)))
+                ,constant-expr)))
+  (def-constantly-fun constantly-t t)
+  (def-constantly-fun constantly-nil nil)
+  (def-constantly-fun constantly-0 0))
+
 (in-package "SB-KERNEL")
 
 (defun fp-zero-p (x)
@@ -1410,15 +1422,6 @@ NOTE: This interface is experimental and subject to change."
              (sorted (stable-sort wrapped comparator :key #'cdr)))
         (map-into sorted #'car sorted))))
 
-;;; Ensure basicness if possible, and simplicity always
-(defun possibly-base-stringize (s)
-  (declare (string s))
-  (cond #+(and sb-unicode (not sb-xc-host))
-        ((and (typep s '(array character (*))) (every #'base-char-p s))
-         (coerce s 'base-string))
-        (t
-         (coerce s 'simple-string))))
-
 ;;; Return T if X is an object that always evaluates to itself. Guard against:
 ;;;  (LET ((S 'XXX)) (SET S 9) (UNINTERN S) (IMPORT S 'KEYWORD) (SYMBOL-VALUE S)) => 9
 ;;; whereby :XXX satisfies KEWORDP and has value 9, or maybe even has no value
