@@ -747,21 +747,45 @@ functions when called with no arguments."
 
 (defun set-tracing-bit (code bit)
   (declare (ignorable code bit))
-  #+64-bit ; there are no bits to spare in a 32-bit header word
+  ;; there are no bits to spare in a 32-bit header word,
+  ;; and with darwin-jit it's not worth the extra complexity
+  #+(and 64-bit (not darwin-jit))
   (with-pinned-objects (code)
     (let ((sap (int-sap (get-lisp-obj-address code))))
       ;; NB: This is not threadsafe on machines that don't promise that
       ;; stores to single bytes are atomic.
-      (setf (sb-vm::sap-ref-8-jit sap #+little-endian (- 1 sb-vm:other-pointer-lowtag)
-                                      #+big-endian (- 6 sb-vm:other-pointer-lowtag))
+      (setf (sap-ref-8 sap #+little-endian (- 2 sb-vm:other-pointer-lowtag)
+                           #+big-endian (- 5 sb-vm:other-pointer-lowtag))
             bit)
-      ;; touch the card mark
+      ;; touch the card mark - WHY???
       (setf (code-header-ref code 1) (code-header-ref code 1)))))
 
 ;;; FIXME: Symbol is lost by accident
 (eval-when (:compile-toplevel :load-toplevel)
   (export 'sb-int::encapsulate-funobj 'sb-int))
 
+;;; Suppose you want to trace function #'FOO no matter how a caller
+;;; references it (maybe capturing #'FOO in a variable before asking
+;;; to trace FOO). We can do that without resorting to breakpoints,
+;;; by replacing the simple-fun entry point in the header of the code
+;;; that contains FOO such that it points to a different simple-fun
+;;; outside of itself. That other simple-fun calls the tracing routine
+;;; and then the real FOO. An entry point can't be replaced with a
+;;; closure, because CLOSURE and SIMPLE-FUN are not fungible.
+;;;
+;;;   +-------------------------+        +--------------------+
+;;;   | codeblob foo            |        | codeblob "TRACER"  |
+;;;   | ...                     |        |                    |
+;;;   | ... boxed data ...      |        | boxed word: #'foo  | -> the "real" FOO
+;;;   | ...                     |        |                    |
+;;;   | ... unboxed data ...    |        |                    |
+;;;   | ...                     |        +--------------------+
+;;;   | simple-fun-header #'foo |    --> | call trace helper  |
+;;;   | redirected entry point  | --/    +--------------------+
+;;;   | instructions of #'FOO   |
+;;;   | ...                     |
+;;;   +-------------------------+
+
 (defun compile-funobj-encapsulation (wrapper info actual-fun)
   #+(or x86 x86-64)
   (let ((code
@@ -800,7 +824,7 @@ functions when called with no arguments."
 ;;; In contrast, ENCAPSULATE-FUNOBJ encapsulates TRACED-FUN by changing the
 ;;; entry point of the function to redirect to a tracing wrapper which then
 ;;; calls back to the correct entry point.
-(defun encapsulate-funobj (traced-fun &optional fdefn)
+(defun encapsulate-funobj (traced-fun &optional (name nil namep))
   (declare (type (or simple-fun closure) traced-fun))
   (let* ((proxy-fun
            (typecase traced-fun
@@ -812,14 +836,13 @@ functions when called with no arguments."
               #+(or x86-64 arm64)
               (with-pinned-objects ((%closure-fun traced-fun))
                 (sb-vm::%alloc-closure 0 (sb-vm::%closure-callee traced-fun)))
-              #-(or x86-64 arm64) (%primitive sb-vm::make-closure traced-fun nil 0 nil))
+              #-(or x86-64 arm64) (%primitive sb-vm::make-closure traced-fun 0 nil))
              (closure
               ;; Same as above, but simpler - the original closure will redirect
               ;; to the tracing wraper, which will invoke a new closure that is
               ;; behaviorally identical to the original closure.
               (sb-impl::copy-closure traced-fun))))
-         (info (make-trace-info :what (cond (fdefn (fdefn-name fdefn))
-                                            (t (%fun-name traced-fun)))
+         (info (make-trace-info :what (if namep name (%fun-name traced-fun))
                                 :encapsulated t
                                 :report 'trace))
          (tracing-wrapper
@@ -857,7 +880,7 @@ functions when called with no arguments."
              (setf (sap-ref-lispobj (int-sap (get-lisp-obj-address traced-fun))
                                     (- sb-vm:n-word-bytes sb-vm:fun-pointer-lowtag))
                    tracing-wrapper))))))
-    ;; Update fdefn's raw-addr slot to point to the tracing wrapper
-    (when (and fdefn (eq (fdefn-fun fdefn) traced-fun))
-      (setf (fdefn-fun fdefn) tracing-wrapper))
+    ;; Possibly update #'NAME to point to the tracing wrapper
+    (when (and namep (eq (fboundp name) traced-fun))
+      (setf (fdefn-fun (find-fdefn name)) tracing-wrapper))
     tracing-wrapper))
