@@ -58,30 +58,33 @@
 ;;;; macros and functions for character tables
 
 ;;; As efficiently as possible look up the character attributes of CHAR.
-;;; If CAST is true then insert an assertion that CHARACTER is actually
-;;; a character. This seems to be necessary to ensure that gray stream
-;;; methods that should return characters don't return random objects.
-(defmacro char-syntax (char syntax-array extension-table &optional cast)
-  #-sb-unicode (declare (ignore cast))
-  `(progn
-     #-sb-unicode ,extension-table
-     (truly-the (unsigned-byte 8)
-                #-sb-unicode
-                (elt ,syntax-array (char-code ,char))
-                #+sb-unicode
-                (if (typep ,char 'base-char)
-                    (elt ,syntax-array (char-code ,char))
-                    (car (truly-the list
-                                    (gethash ,(if cast `(the character ,char) char)
-                                             ,extension-table
-                                             ',`(,+char-attr-constituent+))))))))
+;;; If TYPECHECK then insert an assertion that CHAR is actually a character,
+;;; which is needed sometimes when calling Gray stream methods that could return
+;;; random objects which must be of type CHARACTER.
+;;; (Naturally GETHASH doesn't care what its argument is.)
+(defmacro char-syntax (char syntax-array extension-table &optional typecheck)
+  (declare (ignorable extension-table typecheck))
+  `(truly-the (unsigned-byte 8)
+      ;; Unfortunately we require divergent code here for unicode/non-unicode builds.
+      ;; Non-unicode will deduce that the 'else' branch of the IF can't possibly satisfy
+      ;; the typecheck in `(THE CHARACTER) because CHARACTER = BASE-CHAR, and complains.
+      #+sb-unicode
+      (if (typep ,char 'base-char)
+          (elt ,syntax-array (char-code ,char))
+          (let ((found (gethash ,char ,extension-table +char-attr-constituent+)))
+            (cond ((fixnump found)
+                   ,@(if typecheck `((the character ,char))) ; a "drive-by" typecheck
+                   found)
+                  (t
+                   (car (truly-the cons found))))))
+      #-sb-unicode (elt ,syntax-array (char-code ,char))))
 
 ;;; Formerly known as SET-CAT-ENTRY and SET-CMT-ENTRY
 (defun assign-char-syntax (readtable char attributes function)
   (declare (readtable readtable)
            (character char)
            (type (unsigned-byte 8) attributes)
-           (type (or null function fdefn) function))
+           (type (or function symbol) function))
   (if (typep char 'base-char)
       (setf (aref (base-char-syntax-array readtable) (char-code char)) attributes
             (aref (base-char-macro-array readtable) (char-code char)) function)
@@ -95,10 +98,9 @@
                (remhash char table)))))
   nil)
 
-;;; the value actually stored in the character macro table. As per
-;;; ANSI #'GET-MACRO-CHARACTER and #'SET-MACRO-CHARACTER, this can
-;;; be either a function-designator or NIL, except that we store
-;;; symbols not as themselves but as their #<fdefn>.
+;;; Get the value stored in the character macro table for CHAR. As per
+;;; ANSI #'GET-MACRO-CHARACTER and #'SET-MACRO-CHARACTER, this return as
+;;; function-designator (possibly NIL).
 (defun get-raw-cmt-entry (char readtable)
   (declare (character char) (readtable readtable))
   (if (typep char 'base-char)
@@ -106,8 +108,7 @@
       ;; extended-char entry if present is a cons of the attributes and function
       (cdr (truly-the list (gethash char (extended-char-table readtable))))))
 
-;; As above but get the entry for SUB-CHAR in a dispatching macro table.
-
+;;; As above but get the entry for SUB-CHAR in a dispatching macro table.
 (defmacro charmacro-dtable-base-chars (dtable)
   `(truly-the (simple-vector ,base-char-code-limit) (car ,dtable)))
 (defmacro charmacro-dtable-extended-chars (dtable)
@@ -115,32 +116,22 @@
 (defun get-raw-cmt-dispatch-entry (sub-char sub-table)
   (declare (character sub-char))
   (if (typep sub-char 'base-char)
-      (svref (charmacro-dtable-base-chars (truly-the cons sub-table)) (char-code sub-char))
+      (svref (charmacro-dtable-base-chars (truly-the cons sub-table))
+             (char-code (char-upcase (truly-the base-char sub-char))))
       (awhen (charmacro-dtable-extended-chars sub-table)
-        (gethash sub-char it))))
+        (gethash (char-upcase sub-char) it))))
 
-;; Coerce THING to a character-macro-table entry
-(defmacro coerce-to-cmt-entry (thing)
-  `(let ((x ,thing))
-     (if (typep x '(or null function)) x (find-or-create-fdefn x))))
-
-;; Return a callable function given a character-macro-table entry.
-(defmacro cmt-entry-to-function (val fallback)
-  `(let ((x ,val))
-     (truly-the
-      function
-      (cond ((functionp x) x)
-            ((null x) ,fallback)
-            (t (sb-c:safe-fdefn-fun x))))))
-
-;; Return a function-designator given a character-macro-table entry.
-(defmacro cmt-entry-to-fun-designator (val)
-  `(let ((x ,val))
-     (if (fdefn-p x) (fdefn-name x) x)))
+(eval-when (:compile-toplevel) ; wire in the fallbacks via :KNOWN-FUN constants
+  (sb-c:defknown read-token (stream character) * () :overwrite-fndb-silently t)
+  (sb-c:defknown dispatch-char-error (stream character t) * () :overwrite-fndb-silently t))
+;;; Invoke the character macro table entry for FUN-DESIGNATOR passing it
+;;; ARGS, but if DESIGNATOR is NIL then call FALLBACK instead.
+(defmacro invoke-cmt-entry ((fun-designator fallback) &rest args)
+  `(funcall (truly-the (or function symbol) (or ,fun-designator ,fallback))
+            ,@args))
 
 ;;; The character attribute table is a BASE-CHAR-CODE-LIMIT vector
 ;;; of (unsigned-byte 8) plus a hashtable to handle higher character codes.
-
 (defmacro test-attribute (char whichclass readtable &optional cast)
   (let ((temp '#:readtable))
     `(let ((,temp ,readtable))
@@ -157,15 +148,10 @@
 (declaim (inline single-escape-p multiple-escape-p))
 (declaim (inline token-delimiterp))
 
-;;; the [1] and [2] here refer to ANSI glossary entries for
-;;; "whitespace".
-;; whitespace[2]p is the only predicate whose readtable is optional
-;; - other than whitespace[1]p which has a fixed readtable - due to
-;; callers not otherwise needing a readtable at all, and so not binding
-;; *READTABLE* into a local variable throughout their lifetime.
+;;; the [1] and [2] here refer to ANSI glossary entries for "whitespace".
 (defun whitespace[1]p (char)
   (test-attribute char +char-attr-whitespace+ *standard-readtable*))
-(defun whitespace[2]p (char &optional (rt *readtable*))
+(defun whitespace[2]p (char rt)
   (test-attribute char +char-attr-whitespace+ rt t))
 
 (defun constituentp (char rt)
@@ -302,8 +288,7 @@ be interned (returned, respectively) as required. The default is :SYMBOLS."
        (find-if-in-closure #'consp fun)))
 
 ;;; Copy a char-macro-table entry.
-;;; If ENTRY is a dispatching macro, copy its dispatch table.
-;;; Otherwise return it unchanged, be it NIL or a function or fdefn.
+;;; If ENTRY is a dispatching macro, copy its dispatch table; otherwise return it unchanged.
 (defun copy-cmt-entry (entry)
   (let ((dtable (%dispatch-macro-char-table entry)))
     (if dtable
@@ -376,7 +361,7 @@ standard Lisp readtable when NIL."
      designated-readtable
      char
      (if non-terminatingp +char-attr-constituent+ +char-attr-terminating-macro+)
-     (coerce-to-cmt-entry function))
+     function)
     t)) ; (ANSI-specified return value)
 
 (defun get-macro-character (char &optional (rt-designator *readtable*))
@@ -387,8 +372,7 @@ standard Lisp readtable when NIL."
   (let* ((designated-readtable (or rt-designator *standard-readtable*))
          ;; the first return value: (OR FUNCTION SYMBOL) if CHAR is a macro
          ;; character, or NIL otherwise
-         (fun-value (cmt-entry-to-fun-designator
-                     (get-raw-cmt-entry char designated-readtable))))
+         (fun-value (get-raw-cmt-entry char designated-readtable)))
     (values fun-value
             ;; NON-TERMINATING-P return value:
             (if fun-value
@@ -436,8 +420,7 @@ standard Lisp readtable when NIL."
     (assert-not-standard-readtable readtable 'set-dispatch-macro-character)
     (when (digit-char-p sub-char)
       (error "SUB-CHAR must not be a decimal digit: ~S" sub-char))
-    (let ((dtable (get-dispatch-macro-char-table disp-char readtable))
-          (function (coerce-to-cmt-entry function)))
+    (let ((dtable (get-dispatch-macro-char-table disp-char readtable)))
       ;; (SET-MACRO-CHARACTER #\$ (GET-MACRO-CHARACTER #\#)) will share
       ;; the dispatch table. Perhaps it should be copy-on-write?
       (if (typep sub-char 'base-char)
@@ -458,20 +441,18 @@ standard Lisp readtable when NIL."
    or NIL if there is no associated function."
   (let ((dtable (get-dispatch-macro-char-table
                  disp-char (or rt-designator *standard-readtable*))))
-    (cmt-entry-to-fun-designator
-     (get-raw-cmt-dispatch-entry (char-upcase sub-char) dtable))))
+    (get-raw-cmt-dispatch-entry sub-char dtable)))
 
 
 ;;;; definitions to support internal programming conventions
 
 (defconstant +EOF+ 0)
 
-(defun flush-whitespace (stream)
+(defun flush-whitespace (stream readtable)
   ;; This flushes whitespace chars, returning the last char it read (a
   ;; non-white one). It always gets an error on end-of-file.
-  (let* ((rt *readtable*)
-         (base (base-char-syntax-array rt))
-         (extended (extended-char-table rt)))
+  (let* ((base (base-char-syntax-array readtable))
+         #+sb-unicode (extended (extended-char-table readtable)))
     (macrolet ((done-p ()
                 '(neq (char-syntax char base extended) +char-attr-whitespace+)))
       (if (ansi-stream-p stream)
@@ -522,17 +503,26 @@ standard Lisp readtable when NIL."
 
 ;;;; implementation of the read buffer
 
+(defmacro inline-alloc-string-displaced-to (underlying)
+  `(let ((s (make-array-header #+sb-unicode sb-vm:complex-character-string-widetag
+                               #-sb-unicode sb-vm:complex-base-string-widetag
+                               1)))
+     (setf  (%array-displaced-p s) t
+            (%array-displaced-from s) nil
+            (%array-data s) ,underlying)
+     ;; need of TRULY-THE is understandable, unlike the one below
+     (truly-the (and (array character (*)) (not simple-array)) s)))
+
+;;; This allocator requires 5 pseudo-atomic sections (the structure, two non-simple
+;;; array headers, and two data vectors).
+;;; Too bad we don't have a way to collapse them to one.
 (defstruct (token-buf (:predicate nil) (:copier nil)
-                      (:constructor
-                       make-token-buf
-                       (&aux
-                        (initial-string (make-string 128))
-                        (string initial-string)
-                        (adjustable-string
-                         (make-array 0
-                                     :element-type 'character
-                                     :fill-pointer nil
-                                     :displaced-to string)))))
+                      (:constructor !make-token-buf
+                          (&aux (initial-string (truly-the (simple-array character (128))
+                                                           (make-string 128)))
+                                (string initial-string)
+                                (adjustable-string
+                                 (inline-alloc-string-displaced-to string)))))
   ;; The string accumulated during reading of tokens.
   ;; Always starts out EQ to 'initial-string'.
   (string nil :type (simple-array character (*)))
@@ -616,12 +606,13 @@ standard Lisp readtable when NIL."
 ;; Additionally the cleanup is on a "best effort" basis. Async unwinds
 ;; through WITH-READ-BUFFER fail to recycle token-bufs, but that's ok.
 (defun acquire-token-buf ()
+  (declare (sb-c::tlab :system) (inline !make-token-buf))
   (let ((this-buffer *token-buf-pool*))
     (cond (this-buffer
            (shiftf *token-buf-pool* (token-buf-next this-buffer) nil)
            this-buffer)
           (t
-           (make-token-buf)))))
+           (!make-token-buf)))))
 
 (defun release-token-buf (chain)
   (named-let free ((buffer chain))
@@ -699,17 +690,17 @@ standard Lisp readtable when NIL."
 ;;; Like READ-PRESERVING-WHITESPACE, but doesn't check the read buffer
 ;;; for being set up properly.
 (defun %read-preserving-whitespace (stream eof-error-p eof-value recursive-p)
-  (declare (optimize (sb-c::check-tag-existence 0)))
   (if recursive-p
       ;; a loop for repeating when a macro returns nothing
       (let* ((tracking-p (form-tracking-stream-p stream))
+             (rt *readtable*)
              (outermost-p
               (and tracking-p
                    (null (form-tracking-stream-form-start-char-pos stream)))))
         (loop
          (let ((char (read-char stream eof-error-p +EOF+)))
            (cond ((eq char +EOF+) (return eof-value))
-                 ((whitespace[2]p char))
+                 ((whitespace[2]p char rt))
                  (t
                   (when outermost-p
                     ;; Calling FILE-POSITION at each token seems to slow down
@@ -766,9 +757,8 @@ standard Lisp readtable when NIL."
     (and (form-tracking-stream-p stream)
          ;; Subtract 1 because the position points _after_ CHAR.
          (1- (form-tracking-stream-input-char-pos stream)))
-    (funcall (cmt-entry-to-function
-              (get-raw-cmt-entry char *readtable*) #'read-token)
-             stream char)))
+    (invoke-cmt-entry ((get-raw-cmt-entry char *readtable*) #'read-token)
+                      stream char)))
 
 (defun read (&optional (stream *standard-input*)
                        (eof-error-p t)
@@ -787,7 +777,7 @@ standard Lisp readtable when NIL."
     (unless (or (eql result local-eof-val) recursive-p)
       (let ((next-char (read-char stream nil +EOF+)))
         (unless (or (eq next-char +EOF+)
-                    (whitespace[2]p next-char))
+                    (whitespace[2]p next-char *readtable*))
           (unread-char next-char stream))))
     (if (eq result local-eof-val) eof-value result)))
 
@@ -832,9 +822,10 @@ standard Lisp readtable when NIL."
     ((with-list-reader ((streamvar delimiter) &body body)
        `(let* ((thelist (list nil))
                (listtail thelist)
+               (rt *readtable*)
                (collectp (if *read-suppress* 0 -1)))
           (declare (dynamic-extent thelist))
-          (loop (let ((firstchar (flush-whitespace ,streamvar)))
+          (loop (let ((firstchar (flush-whitespace ,streamvar rt)))
                   (when (eq firstchar ,delimiter)
                     (return (cdr thelist)))
                   ,@body))))
@@ -857,8 +848,8 @@ standard Lisp readtable when NIL."
                         (unless (zerop collectp)
                           (simple-reader-error
                            stream "Nothing appears before . in list.")))
-                       ((whitespace[2]p nextchar)
-                        (setq nextchar (flush-whitespace stream))))
+                       ((whitespace[2]p nextchar rt)
+                        (setq nextchar (flush-whitespace stream rt))))
                  (rplacd (truly-the cons listtail)
                          (read-after-dot stream nextchar collectp))
                  ;; Check for improper ". ,@" or ". ,." now rather than
@@ -896,8 +887,9 @@ standard Lisp readtable when NIL."
 
 (defun read-after-dot (stream firstchar collectp)
   ;; FIRSTCHAR is non-whitespace!
-  (let ((lastobj ()))
-    (do ((char firstchar (flush-whitespace stream)))
+  (let ((lastobj ())
+        (rt *readtable*))
+    (do ((char firstchar (flush-whitespace stream rt)))
         ((eq char #\))
          (if (zerop collectp)
              (return-from read-after-dot nil)
@@ -908,7 +900,7 @@ standard Lisp readtable when NIL."
     ;; At least one thing appears after the dot.
     ;; Check for more than one thing following dot.
     (loop
-     (let ((char (flush-whitespace stream)))
+     (let ((char (flush-whitespace stream rt)))
        (cond ((eq char #\)) (return lastobj)) ;success!
              ;; Try reading virtual whitespace.
              ((not (zerop (logand (read-maybe-nothing stream char)
@@ -1232,7 +1224,7 @@ extended <package-name>::<form-in-package> syntax."
   (let* ((rt *readtable*)
          (base *read-base*)
          (attribute-array (base-char-syntax-array rt))
-         (attribute-hash-table (extended-char-table rt))
+         #+sb-unicode (attribute-hash-table (extended-char-table rt))
          (buf *read-buffer*)
          (package-designator nil)
          (colons 0)
@@ -1821,22 +1813,23 @@ extended <package-name>::<form-in-package> syntax."
 
 (defun read-dispatch-char (stream dispatch-table)
   ;; Read some digits.
-  (let ((numargp nil)
-        (numarg 0)
-        (sub-char ()))
-    (loop
-     (let ((ch (read-char stream nil +EOF+)))
-       (if (eq ch +EOF+)
-           (reader-eof-error stream "inside dispatch character")
-          ;; Take care of the extra char.
-           (let ((dig (digit-char-p ch)))
-             (if dig
-                 (setq numargp t numarg (+ (* numarg 10) dig))
-                 (return (setq sub-char (char-upcase ch))))))))
+  (let* ((numarg nil)
+         (sub-char
+          (loop
+           (let ((ch (read-char stream nil +EOF+)))
+             (if (eq ch +EOF+)
+                 (reader-eof-error stream "inside dispatch character")
+                ;; Take care of the extra char.
+                 (let ((dig (digit-char-p ch)))
+                   (cond ((not dig) (return ch))
+                         ((not numarg) (setq numarg dig))
+                         (t (setq numarg (+ (* numarg 10) dig))))))))))
     ;; Look up the function and call it.
-    (let ((fn (get-raw-cmt-dispatch-entry sub-char dispatch-table)))
-      (funcall (cmt-entry-to-function fn #'dispatch-char-error)
-               stream sub-char (if numargp numarg nil)))))
+    ;; We used to give the user the upcased sub-char, but not only is that
+    ;; not stipulated, the "lossiness" could be construed as a bug.
+    (invoke-cmt-entry ((get-raw-cmt-dispatch-entry sub-char dispatch-table)
+                       #'dispatch-char-error)
+                      stream sub-char numarg)))
 
 ;;;; READ-FROM-STRING
 
@@ -1952,8 +1945,6 @@ extended <package-name>::<form-in-package> syntax."
                  (loop for fn across (the simple-vector (charmacro-dtable-base-chars dtable))
                        and ch from 0
                        when fn do (push (cons (code-char ch) fn) output))
-                 (dolist (cell output) ; coerce values to function-designator
-                   (rplacd cell (cmt-entry-to-fun-designator (cdr cell))))
                  (when hash-table-p     ; caller wants hash-tables
                    (setq output (%stuff-hash-table (make-hash-table) output)))
                  (push (cons char output) alist)))))
