@@ -304,7 +304,7 @@ sufficiently motivated to do lengthy fixes."
   #-cheneygc (sb-c::coalesce-debug-info) ; Share even more things
 
   #+sb-fasteval (sb-interpreter::flush-everything)
-  (tune-hashtable-sizes-of-all-packages))
+  (tune-hashset-sizes-of-all-packages))
 
 (defun deinit ()
   (call-hooks "save" *save-hooks*)
@@ -340,17 +340,12 @@ sufficiently motivated to do lengthy fixes."
   ;; Clean up the simulated weak list of covered code components.
   (rplacd sb-c:*code-coverage-info*
           (delete-if-not #'weak-pointer-value (cdr sb-c:*code-coverage-info*)))
-  ;; Clearing the hash caches must be done after coalescing ctype instances
-  ;; because coalescing compares by TYPE= which creates more cache entries.
-  (hashset-rehash sb-kernel::*key-info-hashset* nil)
-  (hashset-rehash sb-kernel::*key-info-list-hashset* nil)
-  (coalesce-ctypes)
+  (sb-kernel::rebuild-ctype-hashsets)
   (drop-all-hash-caches)
   (os-deinit)
   ;; Perform static linkage. Functions become un-statically-linked
   ;; on demand, for TRACE, redefinition, etc.
   #+immobile-code (sb-vm::statically-link-core)
-  (invalidate-fd-streams)
   (finalizers-deinit)
   ;; Try to shrink the pathname cache. It might be largely nulls
   (rebuild-pathname-cache)
@@ -363,126 +358,6 @@ sufficiently motivated to do lengthy fixes."
         - nil + nil ++ nil +++ nil
         /// nil // nil / nil))
 
-;;; Try to produce a unique representative of each ctype in memory as
-;;; compared by TYPE=, redirecting references on to the chosen representative.
-;;; In the base SBCL image this removes about 400 ctypes instances.
-;;; When saving a large application it can (and does) remove thousands more.
-;;; This is actually not about space saving, but reducing non-determinism.
-;;; Because of the random nature of the type caches (using opaque hashes that
-;;; are generated based on memory address) it's totally arbitrary when we create
-;;; new instances of ctypes. Coalescing tries to make it less so.  As to
-;;; reproducibility, the fact that type-hash-value is an unintelligent key
-;;; is a big problem. I can't think of how to easily make it intelligent,
-;;; but it might work to zero them all out, and restore the hash on demand
-;;; (much the way symbol-hash is lazily computed) which ought to be fine
-;;; since all hash caches start out empty.
-;;;
-;;; To make this even more aggressive, it should coalesce "bottom up"
-;;; so that ctypes contained in other ctypes would be uniquified first.
-;;; The algorithm is too naive to do that at present.
-;;;
-;;; Doing too much consing within MAP-ALLOCATED-OBJECTS can lead to heap
-;;; exhaustion (due to inhibited GC), so this takes several passes.
-(defun coalesce-ctypes (&optional verbose)
-  (declare (optimize (sb-c::aref-trapping 0)))
-  (let* ((table (make-hash-table :test 'equal))
-         interned-ctypes
-         referencing-objects)
-    (labels ((interesting-subpart-p (part)
-               ;; Heap objects can point to "dead" stack objects - those
-               ;; from a no-longer-existing stack frame - so only examine
-               ;; outgoing references within the dynamic space.
-               ;; As to why the pointing object didn't die - who knows?
-               (and (eq (heap-allocated-p part) :dynamic)
-                    (typep part 'ctype)
-                    ;; PART is not interesting if it points to an interned
-                    ;; ctype, because that's already a canonical object.
-                    (not (minusp (type-hash-value part)))))
-             (coalesce (type &aux
-                               ;; Deal with ctypes instances whose unparser fails.
-                               (spec (and (not (contains-unknown-type-p type))
-                                          (ignore-errors (type-specifier type)))))
-               ;; There are ctypes that unparse to the same s-expression
-               ;; but are *NOT* TYPE=. Some examples:
-               ;;   classoid LIST  vs UNION-TYPE LIST  = (OR CONS NULL)
-               ;;   classoid FLOAT vs UNION-TYPE FLOAT = (OR SINGLE-FLOAT DOUBLE-FLOAT)
-               ;;   classoid REAL  vs UNION-TYPE REAL  = (OR FLOAT RATIONAL)
-               ;;   classoid RATIO vs INTERSECTION-TYPE RATIO = (AND RATIONAL (NOT INTEGER))
-               (if spec
-                   (dolist (choice (gethash spec table)
-                                   (progn (push type (gethash spec table))
-                                          type))
-                     (when (type= choice type)
-                       (return choice)))
-                   type)))
-      ;; Start by collecting interned types, as well as any object that points
-      ;; to a ctype.
-      ;; Interned ctypes (mostly classoids, but a few others) have the aspect
-      ;; that if two specifiers are equal, then they map to the same internal
-      ;; object. This does not discount the possibility that some other ctype
-      ;; could be EQ to that type, as occurs with array upgrading.
-      (sb-vm:map-allocated-objects
-       (lambda (obj type size)
-         (declare (ignore type size))
-         (when (and (typep obj 'ctype) (minusp (type-hash-value obj)))
-           (push obj interned-ctypes))
-         (macrolet ((examine (form)
-                      ;; when the subpart of OBJ is possibly going
-                      ;; to get coalesced, then record OBJ.
-                      `(when (interesting-subpart-p ,form)
-                         (push obj referencing-objects)
-                         (return-from skip))))
-           ;; Wrap a block named other than NIL since
-           ;; DO-REFERENCED-OBJECTS has several named NIL.
-           (block skip (sb-vm:do-referenced-object (obj examine)))))
-       :all)
-      (when verbose
-        (format t "Found ~d interned types, ~d referencing objects~%"
-                (length interned-ctypes) (length referencing-objects)))
-      (dolist (type interned-ctypes)
-        (setf (gethash (type-specifier type) table) (list type)))
-      (dolist (obj referencing-objects)
-        (let (written)
-          (macrolet ((examine (form &aux (accessor (if (listp form) (car form))))
-                       (cond
-                         ((not (listp form))
-                          ;; do-closure-values passes an access form that
-                          ;; can't be inverted to a writing form
-                          `(progn ,form nil))
-                         ((eq accessor 'data-vector-ref)
-                          `(let ((part ,form))
-                             (when (interesting-subpart-p part)
-                               (let ((new (coalesce part)))
-                                 (unless (eq new part)
-                                   (setf (svref obj ,(caddr form)) new
-                                         written t))))))
-                         ((and (eq accessor '%primitive)
-                               (eq (cadr form) 'sb-c:fast-symbol-global-value))
-                          `(let ((part ,form))
-                             (when (interesting-subpart-p part)
-                               ;; just do it - skip the attempt-to-modify check
-                               (%set-symbol-global-value obj (coalesce part)))))
-                         ((not (memq accessor
-                                     '(%closure-fun
-                                       %fun-layout %instance-layout
-                                       symbol-package symbol-name symbol-%info
-                                       fdefn-name
-                                       %numerator %denominator
-                                       %realpart %imagpart
-                                       %make-lisp-obj ; fdefn referent
-                                       ;; hope no weak pointers point at ctypes
-                                       weak-pointer-value)))
-                          `(let ((part ,form))
-                             (when (interesting-subpart-p part)
-                               (setf ,form (coalesce part))))))))
-            (sb-vm:do-referenced-object (obj examine)
-              (simple-vector
-               :extend
-               (when (and written (test-header-data-bit obj
-                                                        (ash sb-vm:vector-addr-hashing-flag
-                                                             sb-vm:array-flags-data-position)))
-                 (setf (svref obj 1) 1)))))))))) ; set need-to-rehash
-
 sb-c::
 (defun coalesce-debug-info ()
   #+cheneygc (clrhash sb-di::*compiled-debug-funs*)
