@@ -224,20 +224,19 @@
 #+immobile-code
 (progn
 (defconstant trampoline-entry-offset n-word-bytes)
-;;; TODO: a simplifying trampoline should be a funcallable-instance,
-;;; and not a simple-fun-less piece of code.
 (defun make-simplifying-trampoline (fun)
   ;; 'alloc' is compiled after this file so we don't see the derived type.
   ;; But slam found a conflict on recompile.
   (let ((code (truly-the (values code-component (integer 0) &optional)
                          (allocate-code-object :dynamic 3 24)))) ; KLUDGE
     (setf (%code-debug-info code) fun)
-    (let ((sap (sap+ (code-instructions code) trampoline-entry-offset))
-          (ea (+ (logandc2 (get-lisp-obj-address code) lowtag-mask)
-                 (ash code-debug-info-slot word-shift))))
-      (setf (sap-ref-32 sap 0) #x058B48 ; REX MOV [RIP-n]
-            (signed-sap-ref-32 sap 3) (- ea (+ (sap-int sap) 7)); disp
-            (sap-ref-32 sap 7) #xFD60FF)) ; JMP [RAX-3]
+    (with-pinned-objects (code)
+      (let ((sap (sap+ (code-instructions code) trampoline-entry-offset))
+            (ea (+ (logandc2 (get-lisp-obj-address code) lowtag-mask)
+                   (ash code-debug-info-slot word-shift))))
+        (setf (sap-ref-32 sap 0) #x058B48 ; REX MOV [RIP-n]
+              (signed-sap-ref-32 sap 3) (- ea (+ (sap-int sap) 7)); disp
+              (sap-ref-32 sap 7) #xFD60FF))) ; JMP [RAX-3]
     ;; Verify that the jump table size reads as  0.
     (aver (zerop (code-jump-table-words code)))
     ;; It is critical that there be a trailing 'uint16' of 0 in this object
@@ -247,10 +246,6 @@
     (aver (zerop (code-n-entries code)))
     code))
 
-;;; Return T if FUN can't be called without loading RAX with its descriptor.
-(defun fun-requires-simplifying-trampoline-p (fun)
-  (= (%fun-pointer-widetag fun) closure-widetag))
-
 (defun fdefn-has-static-callers (fdefn)
   (declare (type fdefn fdefn))
   (with-pinned-objects (fdefn)
@@ -269,7 +264,7 @@
            (values function))
   (when (fdefn-has-static-callers fdefn)
     (remove-static-links fdefn))
-  (let ((trampoline (when (fun-requires-simplifying-trampoline-p fun)
+  (let ((trampoline (when (closurep fun)
                       (make-simplifying-trampoline fun)))) ; a newly made CODE object
     (with-pinned-objects (fdefn trampoline fun)
       (let* ((jmp-target
@@ -350,18 +345,14 @@
 ;;; MAP-OBJECTS-IN-RANGE, which is unreliable in the presence of
 ;;; multiple threads. Unfortunately, some users dangerously redefine
 ;;; builtin functions, and moreover, while there are multiple threads.
-(defun statically-link-code-obj (code fixups)
-  (declare (ignorable code fixups))
-  (unless (immobile-space-obj-p code)
-    (return-from statically-link-code-obj code))
+(defun statically-link-code-obj (code fixups &optional observable-fdefns)
+  (declare (ignorable code fixups observable-fdefns))
   #+immobile-code
-  (let* ((fdefns-start (+ code-constants-offset
-                          (* code-slots-per-simple-fun (code-n-entries code))))
-         (fdefns-count (the index (code-n-named-calls code)))
-         (replacements (make-array fdefns-count :initial-element nil))
-         (ambiguous (make-array fdefns-count :initial-element 0 :element-type 'bit))
-         (any-replacements)
-         (any-ambiguous))
+  (binding* (((fdefns-start fdefns-count) (code-header-fdefn-range code))
+             (replacements (make-array fdefns-count :initial-element nil))
+             (ambiguous (make-array fdefns-count :initial-element 0 :element-type 'bit))
+             (any-replacements nil)
+             (any-ambiguous nil))
     ;; For each fdefn, decide two things:
     ;; * whether the fdefn can be replaced by its function - possible only when
     ;;   that function is in immobile space and needs no trampoline.
@@ -372,7 +363,7 @@
       (let* ((fdefn (code-header-ref code (+ fdefns-start i)))
              (fun (when (fdefn-p fdefn) (fdefn-fun fdefn))))
         (when (and (immobile-space-obj-p fun)
-                   (not (fun-requires-simplifying-trampoline-p fun))
+                   (not (closurep fun))
                    (not (member (fdefn-name fdefn) *never-statically-link* :test 'equal))
                    (neq (info :function :inlinep (fdefn-name fdefn)) 'notinline))
           (setf any-replacements t (aref replacements i) fun))))
@@ -413,7 +404,7 @@
                        (offset (cdr fixup))
                        (fdefn (code-header-ref code (+ fdefns-start fdefn-index)))
                        (fun (aref replacements fdefn-index)))
-              (when fun
+              (when (and fun (/= (bit ambiguous fdefn-index) 1))
                 ;; Set the statically-linked flag
                 (sb-vm::set-fdefn-has-static-callers fdefn 1)
                 (when (= (bit ambiguous fdefn-index) 1)
@@ -425,6 +416,7 @@
                   (setf (signed-sap-ref-32 insts offset)
                         (sap- entry (sap+ insts (+ offset 4))))))))
           ;; Replace ambiguous elements of the code header while still holding the lock
+          #+statically-link-if-ambiguous ; never enabled
           (dotimes (i fdefns-count)
             (when (= (bit ambiguous i) 1)
               (let ((wordindex (+ fdefns-start i))
@@ -433,6 +425,75 @@
                       (cons (code-header-ref code wordindex) locs)))))))))
   code)
 
+(defmacro static-call-entrypoint-vector ()
+  '(- (get-lisp-obj-address sb-fasl::*asm-routine-vector*)
+      (ash (+ vector-data-offset (align-up (length +static-fdefns+) 2)) word-shift)))
+
+;;; Return either the address to jump to when calling NAME, or the address
+;;; containing the address, depending on FIXUP-KIND.
+;;; Use FDEFINITION because it strips encapsulations - whether that's
+;;; the right behavior for it or not is a separate concern.
+;;; If somebody tries (TRACE LENGTH) for example, it should not cause
+;;; compilations to fail on account of LENGTH becoming a closure.
+(defun function-raw-address (name fixup-kind &aux (fun (fdefinition name)))
+  (declare (type (member :abs32 :rel32) fixup-kind))
+  (cond ((not (immobile-space-obj-p fun))
+         (error "Can't statically link to ~S: code is movable" name))
+        ((neq (%fun-pointer-widetag fun) simple-fun-widetag)
+         (error "Can't statically link to ~S: non-simple function" name))
+        ((eq fixup-kind :rel32)
+         ;; if performing a relative fixup, return where the function really is,
+         ;; given that calling from anywhere in immobile space to immobile space
+         ;; needs only a signed imm32 operand.
+         (sap-ref-word (int-sap (get-lisp-obj-address fun))
+                       (- (ash simple-fun-self-slot word-shift) fun-pointer-lowtag)))
+        #+immobile-space
+        (t
+         ;; if calling from dynamic space, it is emitted as "call [abs]" where the
+         ;; absolute address is in static space. Return the address of the element
+         ;; in the entrypoint vector, not the address of the function.
+         (let ((vector-data (sap+ (vector-sap sb-fasl::*asm-routine-vector*)
+                                  (- (ash (+ vector-data-offset
+                                             (align-up (length +static-fdefns+) 2))
+                                          word-shift))))
+               (index (the (not null) (position name +static-fdefns+))))
+           (the (signed-byte 32)
+                (sap-int (sap+ vector-data (ash index word-shift))))))))
+
+;; Return the address to which to jump when calling FDEFN,
+;; which is either an fdefn or the name of an fdefn.
+(defun fdefn-entry-address (fdefn)
+  (let ((fdefn (if (fdefn-p fdefn) fdefn (find-or-create-fdefn fdefn))))
+    (+ (get-lisp-obj-address fdefn)
+       (- 2 sb-vm:other-pointer-lowtag))))
+
+(defun validate-asm-routine-vector ()
+  ;; If the jump table in static space does not match the jump table
+  ;; in *assembler-routines*, fix the one one in static space.
+  ;; It's OK that this is delayed until startup, because code pertinent to
+  ;; core restart always uses relative jumps to asm code.
+  #+immobile-space
+  (let* ((code sb-fasl:*assembler-routines*)
+         (external-table (truly-the (simple-array word (*))
+                                    sb-fasl::*asm-routine-vector*))
+         (insts (code-instructions code))
+         (n (sb-impl::hash-table-%count (car (%%code-debug-info code)))))
+    (declare (optimize (insert-array-bounds-checks 0)))
+    (dotimes (i n)
+      (unless (= (aref external-table i) 0)
+        (setf (aref external-table i)
+              (sap-ref-word insts (truly-the index (ash (1+ i) word-shift))))))
+    ;; Preceding the asm routine vector is the vector of addresses of static-fdefns [sic].
+    ;; These are functions deemed particularly important, so they can be called using 1 instruction
+    ;; from any address in dynamic space. The fdefns aren't actually in static space.
+    (let ((vector (truly-the (simple-array word (*))
+                             (%make-lisp-obj (static-call-entrypoint-vector)))))
+      (dotimes (i (length sb-vm:+static-fdefns+))
+        (setf (aref vector i)
+              (let ((fun (%symbol-function (truly-the symbol (aref +static-fdefns+ i)))))
+                (sap-ref-word (int-sap (get-lisp-obj-address fun))
+                              (- (ash simple-fun-self-slot word-shift) fun-pointer-lowtag))))))))
+
 (sb-c::when-vop-existsp (:translate sb-c::unsigned+)
   (defconstant cf-bit 0)
   (defconstant sf-bit 7)
