@@ -169,13 +169,15 @@
 (compile 'rename-file-a-la-unix)
 
 (export '(*target-sbcl-version* *generated-sources-root*
+          *build-dependent-generated-sources-root*
           stem-source-path find-bootstrap-file read-from-file))
 (defvar *sources-root* "")
 (defvar *generated-sources-root* "")
+(defvar *build-dependent-generated-sources-root* "")
 (defvar *src-cold-shared-pathname* *load-pathname*)
 
 ;;; See remark in COMPILE-STEM about strings vs. The Common Lisp Way
-(defun find-bootstrap-file (namestring)
+(defun find-bootstrap-file (namestring &optional build-dependent)
   (cond ((char= (char namestring 0) #\^)
          ;; If it starts with a "^" then it means "src/cold/..."
          (let ((this *src-cold-shared-pathname*)
@@ -190,7 +192,9 @@
          ;; in the tree as checked in, or generated by a prior build step.
          (concatenate 'string
                       (if (eql (mismatch "output/" namestring) 7) ; a generated source
-                          *generated-sources-root*
+                          (if build-dependent
+                              *build-dependent-generated-sources-root*
+                              *generated-sources-root*)
                           *sources-root*)
                       namestring))
         (t
@@ -201,9 +205,10 @@
 
 ;;; Return an expression read from the file named NAMESTRING.
 ;;; For user-supplied inputs, protect against more than one expression
-;;; appearing in the file. For in-tree inputs we needn't bother.
-(defun read-from-file (namestring &optional (enforce-single-expr t))
-  (with-open-file (s (find-bootstrap-file namestring))
+;;; appearing in the file.  (Our ^build-order.lisp-expr file has more than
+;;; one expression in it, so we need to be able to not enforce.)
+(defun read-from-file (namestring &key (enforce-single-expr t) build-dependent)
+  (with-open-file (s (find-bootstrap-file namestring build-dependent))
     (let* ((result (read s))
            (eof-result (cons nil nil)))
       (unless enforce-single-expr
@@ -282,12 +287,12 @@
         (when (target-featurep '(:and :sb-thread (:or :linux :freebsd)))
           (pushnew :sb-futex sb-xc:*features*))
         (when (target-featurep :immobile-space)
+          (when (member :sb-thread sb-xc:*features*)
+            (pushnew :system-tlabs sb-xc:*features*))
           (pushnew :compact-instance-header sb-xc:*features*)
           (pushnew :immobile-code sb-xc:*features*))
         (when (target-featurep :64-bit)
           (push :compact-symbol sb-xc:*features*))
-        (when (target-featurep '(:and :sb-thread (:not :win32)))
-          (push :pauseless-threadstart sb-xc:*features*))
         (when (target-featurep '(:and :sb-thread (:or :darwin :openbsd)))
           (push :os-thread-stack sb-xc:*features*))
         (when (target-featurep '(:and :x86 :int4-breakpoints))
@@ -299,7 +304,9 @@
           ;; Just print something and go on with life.
           (setq sb-xc:*features* (remove :int4-breakpoints sb-xc:*features*))
           (warn "Removed :INT4-BREAKPOINTS from target features"))
-        (when (target-featurep '(:or :arm64 :sse4))
+        (when (or (target-featurep :arm64)
+                  (and (target-featurep :x86-64)
+                       (member :sse4 backend-subfeatures)))
           (push :round-float sb-xc:*features*))
         (when (target-featurep '(:and :arm64 :darwin))
           (push :arm-v8.1 backend-subfeatures))
@@ -328,8 +335,6 @@
 (let ((feature-compatibility-tests
        '(("(and sb-thread (not gencgc))"
           ":SB-THREAD requires :GENCGC")
-         ("(and pauseless-threadstart (not sb-thread))"
-          ":PAUSELESS-THREADSTART requires :SB-THREAD")
          ("(and sb-safepoint (not sb-thread))" ":SB-SAFEPOINT requires :SB-THREAD")
          ("(and sb-thread (not (or riscv ppc ppc64 x86 x86-64 arm64)))"
           ":SB-THREAD not supported on selected architecture")
@@ -360,6 +365,10 @@
           ":IMMOBILE-CODE requires :IMMOBILE-SPACE feature")
          ("(and immobile-symbols (not immobile-space))"
           ":IMMOBILE-SYMBOLS requires :IMMOBILE-SPACE feature")
+         ("(and system-tlabs (or (not sb-thread) (not immobile-space)))"
+          ;; I don't think it really reqires immobile-space but I haven't tried
+          ;; it and I don't care to support that.
+          ":SYSTEM-TLABS requires SB-THREAD and IMMOBILE-SPACE")
          ("(and sb-futex (not sb-thread))"
           "Can't enable SB-FUTEX on platforms lacking thread support")
          ;; There is still hope to make multithreading on DragonFly x86-64
@@ -484,8 +493,7 @@
                               (pathname-type (compile-file-pathname stem)))))
         (:target-compile
          (values *target-obj-prefix*
-                 (cond ((find :extra-artifact flags) "")
-                       ((find :assem flags) *target-assem-obj-suffix*)
+                 (cond ((find :assem flags) *target-assem-obj-suffix*)
                        (t *target-obj-suffix*)))))
     (concatenate 'string obj-prefix (stem-remap-target stem) obj-suffix)))
 (compile 'stem-object-path)
@@ -506,7 +514,7 @@
           ;; to produce warnings as a bug workaround.
           (let ((cl:*features* (cons feature cl:*features*))
                 (*readtable* *xc-readtable*))
-            (read-from-file "^build-order.lisp-expr" nil))))
+            (read-from-file "^build-order.lisp-expr" :enforce-single-expr nil))))
     (setf *stems-and-flags* (cons build-phase list)))
   ;; Now check for duplicate stems and bogus flags.
   (let ((stems (make-hash-table :test 'equal)))
@@ -552,7 +560,8 @@
          (tmp-obj
            (concatenate 'string obj
                         (if *compile-for-effect-only* "-scratch" "-tmp")))
-         (compile-file (ecase mode
+         (compilation-fn
+                       (ecase mode
                          (:host-compile
                           #+abcl ; ABCL complains about its own deficiency and then returns T
                           ;; for warnings and failure. "Unable to compile function" is not our problem,
@@ -582,7 +591,7 @@
                                               *target-compile-file*))))
          (trace-file (if (find :trace-file flags) t nil))
          (block-compile (if (find :block-compile flags) t :specified)))
-    (declare (type function compile-file))
+    (declare (type function compilation-fn))
 
     (ensure-directories-exist obj :verbose cl:*compile-print*) ; host's value
 
@@ -629,10 +638,10 @@
            (report-continue-restart (stream)
              (format stream "Continue, using possibly bogus file ~S" obj)))
       (tagbody
-       retry-compile-file
+       retry-compile
          (multiple-value-bind (output-truename warnings-p failure-p)
              (restart-case
-                 (apply compile-file src
+                 (apply compilation-fn src
                         :output-file tmp-obj
                         :block-compile (and
                                         ;; Block compilation was
@@ -655,7 +664,7 @@
                           '(:trace-file t :print t)))
                (recompile ()
                  :report report-recompile-restart
-                 (go retry-compile-file)))
+                 (go retry-compile)))
            (declare (ignore warnings-p))
            (cond ((not output-truename)
                   (error "couldn't compile ~S" src))
@@ -666,7 +675,7 @@
                                   obj)
                          (recompile ()
                            :report report-recompile-restart
-                           (go retry-compile-file))
+                           (go retry-compile))
                          (continue ()
                            :report report-continue-restart
                            (setf failure-p nil)))
@@ -737,13 +746,26 @@
 ;;; Run the cross-compiler on a file in the source directory tree to
 ;;; produce a corresponding file in the target object directory tree.
 (defun target-compile-stem (stem flags)
-  (funcall *in-target-compilation-mode-fn*
+  (let ((system-tlab-p
+         (or (search "src/pcl" stem)
+             (search "src/code/alieneval" stem)
+             (search "src/code/arena" stem)
+             (search "src/code/avltree" stem)
+             (search "src/code/brothertree" stem)
+             (search "src/code/debug" stem)
+             (search "src/code/early-defmethod" stem)
+             (search "src/code/final" stem)
+             (search "src/code/format" stem)
+             (search "src/code/solist" stem))))
+    (funcall *in-target-compilation-mode-fn*
            (lambda ()
-             (progv (list (intern "*SOURCE-NAMESTRING*" "SB-C"))
-                    (list (lpnify-stem stem))
+             (progv (list (intern "*SOURCE-NAMESTRING*" "SB-C")
+                          (intern "*FORCE-SYSTEM-TLAB*" "SB-C"))
+                    (list (lpnify-stem stem)
+                          system-tlab-p)
                (loop
                 (with-simple-restart (recompile "Recompile")
-                  (return (compile-stem stem flags :target-compile))))))))
+                  (return (compile-stem stem flags :target-compile)))))))))
 (compile 'target-compile-stem)
 
 ;;; (This function is not used by the build process, but is intended
