@@ -618,7 +618,10 @@
                      (if (template-conditional-p info) 0 (length rtypes))
                      (template-more-results-type info) "results")
       (check-tn-refs (vop-temps vop) vop t 0 t "temps")
-      (unless (= (length (vop-codegen-info vop))
+      (unless (= (+ (length (vop-codegen-info vop))
+                    (if (typep (template-result-types info) '(cons (eql :conditional)))
+                        -1
+                        0))
                  (template-info-arg-count info))
         (barf "wrong number of codegen info args in ~S" vop))))
   (values))
@@ -1069,6 +1072,15 @@
            (write-string "enclose ")
            (dolist (leaf (enclose-funs node))
              (print-leaf leaf)
+             (write-char #\space)
+             (let* ((entry-fun (functional-entry-fun leaf))
+                    (env (and entry-fun (lambda-environment entry-fun))))
+               (when env
+                 (write-string "{env:")
+                 (dolist (leaf (environment-closure env))
+                   (write-char #\space)
+                   (print-leaf leaf))
+                 (write-string "}")))
              (write-char #\space))))
         (when (and *debug-print-types*
                    (valued-node-p node))
