@@ -67,7 +67,7 @@
                                  alternative-constraints)
   (case operator
     ((eq eql char= two-arg-char-equal
-      > < =)
+      > < = <= >=)
      (when (= (length args) 2)
        (flet ((ok-lvar-p (lvar)
                 (or (ok-lvar-lambda-var lvar constraints)
@@ -97,6 +97,8 @@
                     (case operator
                       (< '>)
                       (> '<)
+                      (<= '>=)
+                      (>= '<=)
                       (t operator))))
             (when constant-x
               (when constant-y
@@ -200,7 +202,9 @@
                                       (and (eq (equality-constraint-operator con)
                                                (case operator
                                                  (< '>)
-                                                 (> '<)))
+                                                 (> '<)
+                                                 (<= '>=)
+                                                 (>= '<=)))
                                            (constraint-eq-p leaf2 vector-length-p2 (constraint-x con))
                                            (constraint-eq-p leaf1 vector-length-p1 (constraint-y con)))))
                            return con))
@@ -248,6 +252,16 @@
          `(when (and constraint
                      (not (constraint-not-p constraint)))
             (setf result nil)
+            t))
+        (inverse-true
+         `(when (and constraint
+                     (constraint-not-p constraint))
+            (setf result t)
+            t))
+        (inverse-false
+         `(when (and constraint
+                     (constraint-not-p constraint))
+            (setf result nil)
             t)))))
 
 (defoptimizer (eql equality-constraint) ((x y))
@@ -259,7 +273,9 @@
      (equality-constraint two-arg-char-equal false)
      (equality-constraint > inverse)
      (equality-constraint < inverse)
-     (equality-constraint = false))
+     (equality-constraint = false)
+     (equality-constraint <= inverse-false)
+     (equality-constraint >= inverse-false))
     result))
 
 (defoptimizer (eq equality-constraint) ((x y))
@@ -271,7 +287,9 @@
      (equality-constraint two-arg-char-equal false)
      (equality-constraint > inverse)
      (equality-constraint < inverse)
-     (equality-constraint = false))
+     (equality-constraint = false)
+     (equality-constraint <= inverse-false)
+     (equality-constraint >= inverse-false))
     result))
 
 (defoptimizer (= equality-constraint) ((x y))
@@ -281,7 +299,9 @@
      (equality-constraint eq true)
      (equality-constraint eql true)
      (equality-constraint > inverse)
-     (equality-constraint < inverse))
+     (equality-constraint < inverse)
+     (equality-constraint <= inverse-false)
+     (equality-constraint >= inverse-false))
     result))
 
 (defoptimizer (char= equality-constraint) ((x y))
@@ -309,7 +329,8 @@
      (equality-constraint < inverse nil)
      (equality-constraint eq inverse)
      (equality-constraint eql inverse)
-     (equality-constraint = inverse))
+     (equality-constraint = inverse)
+     (equality-constraint <= inverse-true nil))
     result))
 
 (defoptimizer (< equality-constraint) ((x y))
@@ -319,7 +340,26 @@
      (equality-constraint > inverse nil)
      (equality-constraint eq inverse)
      (equality-constraint eql inverse)
-     (equality-constraint = inverse))
+     (equality-constraint = inverse)
+     (equality-constraint >= inverse-true nil))
+    result))
+
+(defoptimizer (>= equality-constraint) ((x y))
+  (let ((result :give-up))
+    (or
+     (equality-constraint >= nil nil)
+     (equality-constraint > true nil)
+     (equality-constraint = true)
+     (equality-constraint < inverse-true nil))
+    result))
+
+(defoptimizer (<= equality-constraint) ((x y))
+  (let ((result :give-up))
+    (or
+     (equality-constraint < true nil)
+     (equality-constraint <= nil nil)
+     (equality-constraint = true)
+     (equality-constraint > inverse-true nil))
     result))
 
 (defoptimizer (- equality-constraint) ((x y) node)
