@@ -274,10 +274,14 @@
   (integer integer) rational
   (no-verify-arg-count unsafely-flushable))
 
-(defknown (two-arg-< two-arg-= two-arg->)
+(defknown (two-arg-< two-arg-= two-arg-> two-arg-<= two-arg->=)
     (number number) boolean
     (no-verify-arg-count))
 
+(defknown (range< range<= range<<= range<=<)
+    (fixnum real fixnum) boolean
+    (foldable flushable movable no-verify-arg-count))
+
 (defknown (two-arg-gcd two-arg-lcm two-arg-and two-arg-ior two-arg-xor two-arg-eqv)
   (integer integer) integer
   (no-verify-arg-count))
@@ -335,7 +339,10 @@
 (defknown (rationalize) (real) rational
   (movable foldable flushable recursive))
 
-(defknown (numerator denominator) (rational) integer
+(defknown numerator (rational) integer
+  (movable foldable flushable))
+
+(defknown denominator (rational) (integer 1)
   (movable foldable flushable))
 
 (defknown (floor ceiling round)
@@ -368,6 +375,13 @@
 (defknown %unary-truncate-double-float-to-bignum (double-float) bignum
    (#+(or) foldable movable flushable fixed-args))
 
+(defknown (subtract-bignum add-bignums) (bignum bignum) integer
+    (movable flushable no-verify-arg-count))
+(defknown (add-bignum-fixnum subtract-bignum-fixnum) (bignum fixnum) integer
+    (movable flushable no-verify-arg-count))
+(defknown subtract-fixnum-bignum (fixnum bignum) integer
+    (movable flushable no-verify-arg-count))
+
 (defknown sxhash-bignum-double-float (double-float) hash-code
   (foldable movable flushable fixed-args))
 (defknown sxhash-bignum-single-float (single-float) hash-code
