@@ -222,6 +222,9 @@
     *immobile-codeblob-tree* ; for generations 0 through 5 inclusive
     *immobile-codeblob-vector* ; for pseudo-static-generation
     *dynspace-codeblob-tree*
+    sb-impl::**finalizer-store**
+    sb-impl::*finalizer-rehashlist*
+    sb-impl::*finalizers-triggered*
 
     ;; stack pointers
     #-sb-thread *binding-stack-start* ; a thread slot if #+sb-thread
