@@ -117,26 +117,15 @@ RETURN-FROM can be used to exit the form."
                                   :cleanup cleanup)))
       (ir1-convert-progn-body dummy next result forms))))
 
+;;; We make NEXT start a block just so that it will have a block
+;;; assigned. People assume that when they pass a ctran into
+;;; IR1-CONVERT as NEXT, it will have a block when it is done.
 (def-ir1-translator return-from ((name &optional value) start next result)
   "RETURN-FROM name value
 
 Evaluate the VALUE, returning its values from the lexically enclosing
 block NAME. This is constrained to be used only within the dynamic
 extent of the block."
-  ;; old comment:
-  ;;   We make NEXT start a block just so that it will have a block
-  ;;   assigned. People assume that when they pass a ctran into
-  ;;   IR1-CONVERT as NEXT, it will have a block when it is done.
-  ;; KLUDGE: Note that this block is basically fictitious. In the code
-  ;;   (BLOCK B (RETURN-FROM B) (SETQ X 3))
-  ;; it's the block which answers the question "which block is
-  ;; the (SETQ X 3) in?" when the right answer is that (SETQ X 3) is
-  ;; dead code and so doesn't really have a block at all. The existence
-  ;; of this block, and that way that it doesn't explicitly say
-  ;; "I'm actually nowhere at all" makes some logic (e.g.
-  ;; BLOCK-HOME-LAMBDA-OR-NULL) more obscure, and it might be better
-  ;; to get rid of it, perhaps using a special placeholder value
-  ;; to indicate the orphanedness of the code.
   (ctran-starts-block next)
   (let* ((found (or (lexenv-find name blocks)
                     (compiler-error "return for unknown block: ~S" name)))
@@ -152,9 +141,6 @@ extent of the block."
     (ir1-convert start value-ctran value-lvar value)
     (push exit (entry-exits entry))
     (link-node-to-previous-ctran exit value-ctran)
-    (let ((home-lambda (ctran-home-lambda-or-null start)))
-      (when home-lambda
-        (sset-adjoin entry (lambda-calls-or-closes home-lambda))))
     (use-continuation exit exit-ctran (third found))))
 
 ;;; Return a list of the segments of a TAGBODY. Each segment looks
@@ -244,9 +230,6 @@ constrained to be used only within the dynamic extent of the TAGBODY."
          (exit (make-exit :entry entry)))
     (push exit (entry-exits entry))
     (link-node-to-previous-ctran exit start)
-    (let ((home-lambda (ctran-home-lambda-or-null start)))
-      (when home-lambda
-        (sset-adjoin entry (lambda-calls-or-closes home-lambda))))
     (use-ctran exit (second found))))
 
 ;;;; translators for compiler-magic special forms
@@ -1240,9 +1223,6 @@ care."
            (when (constant-p leaf)
              (compiler-error "~S is a constant and thus can't be set." name))
            (when (lambda-var-p leaf)
-             (let ((home-lambda (ctran-home-lambda-or-null start)))
-               (when (and home-lambda (neq (lambda-var-home leaf) home-lambda))
-                 (sset-adjoin leaf (lambda-calls-or-closes home-lambda))))
              (when (lambda-var-ignorep leaf)
                ;; ANSI's definition of "Declaration IGNORE, IGNORABLE"
                ;; requires that this be a STYLE-WARNING, not a full warning.
