@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 (defconstant arg-count-sc (make-sc+offset immediate-arg-scn nargs-offset))
 (defconstant closure-sc (make-sc+offset descriptor-reg-sc-number lexenv-offset))
@@ -39,7 +39,7 @@
 ;;; them at a known location.
 (defun make-old-fp-save-location (env)
   (specify-save-tn
-   (physenv-debug-live-tn (make-normal-tn *fixnum-primitive-type*) env)
+   (environment-debug-live-tn (make-normal-tn *fixnum-primitive-type*) env)
    (make-wired-tn *fixnum-primitive-type*
                   control-stack-arg-scn
                   ocfp-save-offset)))
@@ -47,7 +47,7 @@
 (defun make-return-pc-save-location (env)
   (let ((ptype *backend-t-primitive-type*))
     (specify-save-tn
-     (physenv-debug-live-tn (make-normal-tn ptype) env)
+     (environment-debug-live-tn (make-normal-tn ptype) env)
      (make-wired-tn ptype control-stack-arg-scn lra-save-offset))))
 
 ;;; Make a TN for the standard argument count passing location.  We only
@@ -119,8 +119,7 @@
     (emit-label start-lab)
     ;; Allocate function header.
     (inst simple-fun-header-word)
-    (dotimes (i (1- simple-fun-code-offset))
-      (inst word 0))
+    (inst .skip (* (1- simple-fun-insts-offset) n-word-bytes))
     ;; The start of the actual code.
     ;; Compute CODE from the address of this entry point.
     (let ((entry-point (gen-label)))
@@ -149,7 +148,7 @@
     (move res csp-tn)
     (inst addu csp-tn csp-tn
           (* n-word-bytes (sb-allocated-size 'control-stack)))
-    (when (ir2-physenv-number-stack-p callee)
+    (when (ir2-environment-number-stack-p callee)
       (inst addu nsp-tn nsp-tn
             (- (bytes-needed-for-non-descriptor-stack-frame)))
       (move nfp nsp-tn))))
@@ -245,7 +244,7 @@ default-value-8
         ;; gets confused.
         (without-scheduling ()
           (note-this-location vop :single-value-return)
-          (move csp-tn ocfp-tn t)
+          (emit-nop-or-move csp-tn ocfp-tn)
           (inst nop))
         (when lra-label
           (inst compute-code-from-lra code-tn code-tn lra-label temp)))
@@ -260,7 +259,7 @@ default-value-8
           ;; If there are no stack results, clear the stack now.
           (if (> nvals register-arg-count)
               (inst addu temp nargs-tn (fixnumize (- register-arg-count)))
-              (move csp-tn ocfp-tn t)))
+              (emit-nop-or-move csp-tn ocfp-tn)))
 
         ;; Do the single value case.
         (do ((i 1 (1+ i))
@@ -269,7 +268,7 @@ default-value-8
           (move (tn-ref-tn val) null-tn))
         (when (> nvals register-arg-count)
           (inst b default-stack-vals)
-          (move ocfp-tn csp-tn t))
+          (emit-nop-or-move ocfp-tn csp-tn))
 
         (emit-label regs-defaulted)
 
@@ -298,7 +297,7 @@ default-value-8
 
             (let ((defaults (defaults)))
               (aver defaults)
-              (assemble (*elsewhere*)
+              (assemble (:elsewhere)
                 (emit-label default-stack-vals)
                 (do ((remaining defaults (cdr remaining)))
                     ((null remaining))
@@ -349,7 +348,7 @@ default-value-8
 
     (emit-label done)
 
-    (assemble (*elsewhere*)
+    (assemble (:elsewhere)
       (emit-label variable-values)
       (when lra-label
         (inst compute-code-from-lra code-tn code-tn lra-label temp))
@@ -359,7 +358,7 @@ default-value-8
         (storew (first arg) args i))
       (move start args)
       (inst b done)
-      (move count nargs t)))
+      (emit-nop-or-move count nargs)))
   (values))
 
 
@@ -375,8 +374,7 @@ default-value-8
               values-start)
   (:temporary (:sc any-reg :offset nargs-offset
                :from :eval :to (:result 1))
-              nvals)
-  (:temporary (:scs (non-descriptor-reg)) temp))
+              nvals))
 
 
 ;;; This hook in the codegen pass lets us insert code before fall-thru entry
@@ -553,7 +551,7 @@ default-value-8
               (bytes-needed-for-non-descriptor-stack-frame))))
     (inst addu lip return-pc-temp (- n-word-bytes other-pointer-lowtag))
     (inst j lip)
-    (move cfp-tn ocfp-temp t)))
+    (emit-nop-or-move cfp-tn ocfp-temp)))
 
 
 ;;;; Full call:
@@ -616,7 +614,7 @@ default-value-8
           '((ocfp :target ocfp-pass)
             (return-pc :target return-pc-pass)))
 
-      ,@(unless variable '((args :more t :scs (descriptor-reg)))))
+      ,@(unless variable '((args :more t :scs (descriptor-reg control-stack)))))
 
      ,@(when (eq return :fixed)
          '((:results (values :more t))))
@@ -727,7 +725,7 @@ default-value-8
                             '((:load-ocfp
                                (sc-case ocfp
                                  (any-reg
-                                  (move ocfp-pass ocfp t))
+                                  (emit-nop-or-move ocfp-pass ocfp))
                                  (control-stack
                                   (inst lw ocfp-pass cfp-tn
                                         (ash (tn-offset ocfp)
@@ -735,7 +733,7 @@ default-value-8
                               (:load-return-pc
                                (sc-case return-pc
                                  (descriptor-reg
-                                  (move return-pc-pass return-pc t))
+                                  (emit-nop-or-move return-pc-pass return-pc))
                                  (control-stack
                                   (inst lw return-pc-pass cfp-tn
                                         (ash (tn-offset return-pc)
@@ -749,7 +747,7 @@ default-value-8
                               (:frob-nfp
                                (store-stack-tn nfp-save cur-nfp))
                               (:save-fp
-                               (move ocfp-pass cfp-tn t))
+                               (emit-nop-or-move ocfp-pass cfp-tn))
                               (:load-fp
                                ,(if variable
                                     '(move cfp-tn new-fp)
@@ -761,13 +759,13 @@ default-value-8
                 (insert-step-instrumenting (callable-tn)
                   ;; Conditionally insert a conditional trap:
                   (when step-instrumenting
-                    (load-symbol-value stepping sb!impl::*stepping*)
+                    (load-symbol-value stepping sb-impl::*stepping*)
                     ;; If it's not 0, trap.
                     (inst beq stepping STEP-DONE-LABEL)
                     (inst nop)
                     ;; CONTEXT-PC will be pointing here when the
                     ;; interrupt is handled, not after the BREAK.
-                    (note-this-location vop :step-before-vop)
+                    (note-this-location vop :internal-error)
                     (inst break (reg-tn-encoding callable-tn)
                           single-step-around-trap)
                     (emit-label step-done-label))))
@@ -778,11 +776,11 @@ default-value-8
                     (descriptor-reg (move name-pass name))
                     (control-stack
                      (inst lw name-pass cfp-tn
-                           (ash (tn-offset name) word-shift))
+                           (tn-byte-offset name))
                      (do-next-filler))
                     (constant
                      (inst lw name-pass code-tn
-                           (- (ash (tn-offset name) word-shift)
+                           (- (tn-byte-offset name)
                               other-pointer-lowtag))
                      (do-next-filler)))
                   ;; The step instrumenting must be done after
@@ -798,11 +796,11 @@ default-value-8
                     (descriptor-reg (move lexenv arg-fun))
                     (control-stack
                      (inst lw lexenv cfp-tn
-                           (ash (tn-offset arg-fun) word-shift))
+                           (tn-byte-offset arg-fun))
                      (do-next-filler))
                     (constant
                      (inst lw lexenv code-tn
-                           (- (ash (tn-offset arg-fun) word-shift)
+                           (- (tn-byte-offset arg-fun)
                               other-pointer-lowtag))
                      (do-next-filler)))
                   (inst lw function lexenv
@@ -814,7 +812,7 @@ default-value-8
                   ;; is calculated.
                   (insert-step-instrumenting function)
                   (inst addu entry-point function
-                        (- (ash simple-fun-code-offset word-shift)
+                        (- (ash simple-fun-insts-offset word-shift)
                            fun-pointer-lowtag))))
                (:direct
                 `((inst lw entry-point null-tn (static-fun-offset fun)))))
@@ -873,6 +871,7 @@ default-value-8
   (:temporary (:sc any-reg :offset lexenv-offset :from (:argument 1)) lexenv)
   (:temporary (:sc any-reg :offset ocfp-offset :from (:argument 2)) ocfp)
   (:temporary (:sc any-reg :offset lra-offset :from (:argument 3)) lra)
+  (:temporary (:sc any-reg :offset nl3-offset) temp)
 
   (:vop-var vop)
 
@@ -886,7 +885,8 @@ default-value-8
 
     ;; Clear the number stack if anything is there and jump to the
     ;; assembly-routine that does the bliting.
-    (inst j (make-fixup 'tail-call-variable :assembly-routine))
+    (inst li temp (make-fixup 'tail-call-variable :assembly-routine))
+    (inst j temp)
     (let ((cur-nfp (current-nfp-tn vop)))
       (if cur-nfp
         (inst addu nsp-tn cur-nfp
@@ -990,12 +990,12 @@ default-value-8
   (:temporary (:sc any-reg :offset nl0-offset :from (:argument 2)) vals)
   (:temporary (:sc any-reg :offset nargs-offset :from (:argument 3)) nvals)
   (:temporary (:sc descriptor-reg :offset a0-offset) a0)
+  (:temporary (:sc any-reg :offset nl3-offset) temp)
   (:temporary (:scs (interior-reg)) lip)
 
   (:vop-var vop)
 
   (:generator 13
-    (let ((not-single (gen-label)))
       ;; Clear the number stack.
       (let ((cur-nfp (current-nfp-tn vop)))
         (when cur-nfp
@@ -1013,13 +1013,14 @@ default-value-8
       (lisp-return lra-arg lip :offset 2)
 
       ;; Nope, not the single case.
-      (emit-label not-single)
+    NOT-SINGLE
       (move ocfp ocfp-arg)
       (move lra lra-arg)
       (move vals vals-arg)
 
-      (inst j (make-fixup 'return-multiple :assembly-routine))
-      (move nvals nvals-arg t))))
+      (inst li temp (make-fixup 'return-multiple :assembly-routine))
+      (inst j temp)
+      (emit-nop-or-move nvals nvals-arg)))
 
 ;;;; XEP hackery:
 
@@ -1070,7 +1071,7 @@ default-value-8
       ;; Everything of interest in registers.
       (inst blez count do-regs)
       ;; Initialize dst to be end of stack.
-      (move dst csp-tn t)
+      (emit-nop-or-move dst csp-tn)
       ;; Initialize src to be end of args.
       (inst addu src cfp-tn nargs-tn)
 
@@ -1103,17 +1104,55 @@ default-value-8
 ;;; More args are stored consecutively on the stack, starting
 ;;; immediately at the context pointer.  The context pointer is not
 ;;; typed, so the lowtag is 0.
-(define-full-reffer more-arg * 0 0 (descriptor-reg any-reg) * %more-arg)
+(define-vop (more-arg)
+  (:translate %more-arg)
+  (:policy :fast-safe)
+  (:args (context :scs (descriptor-reg)) (index :scs (any-reg)))
+  (:arg-types * tagged-num)
+  (:temporary (:scs (any-reg)) temp)
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:result-types *)
+  (:generator 5
+   (inst addu temp context index)
+   (loadw value temp)))
+(define-vop (more-arg-c)
+  (:translate %more-arg)
+  (:policy :fast-safe)
+  (:args (context :scs (descriptor-reg)))
+  (:info index)
+  (:arg-types * (:constant (load/store-index 8 0 0)))
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:result-types *)
+  (:generator 4
+    (loadw value context index)))
+
+(define-vop (more-arg-or-nil)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg) :to (:result 1))
+         (count :scs (any-reg) :to (:result 1)))
+  (:temporary (:scs (any-reg)) temp)
+  (:info index)
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:result-types *)
+  (:generator 3
+    (cond ((zerop index)
+           (inst beq count done))
+          (t
+           (inst subu temp (fixnumize index) count)
+           (inst bgez temp done)))
+    (move value null-tn)
+    (loadw value object index)
+    done))
 
 ;;; Turn more arg (context, count) into a list.
-(define-vop (listify-rest-args)
+(define-vop ()
   (:args (context-arg :target context :scs (descriptor-reg))
          (count-arg :target count :scs (any-reg)))
   (:arg-types * tagged-num)
   (:temporary (:scs (any-reg) :from (:argument 0)) context)
   (:temporary (:scs (any-reg) :from (:argument 1)) count)
   (:temporary (:scs (descriptor-reg) :from :eval) temp dst)
-  (:temporary (:sc non-descriptor-reg :offset nl4-offset) pa-flag)
+  (:temporary (:sc non-descriptor-reg) pa-flag)
   (:results (result :scs (descriptor-reg)))
   (:translate %listify-rest-args)
   (:policy :safe)
@@ -1122,26 +1161,48 @@ default-value-8
     (let* ((enter (gen-label))
            (loop (gen-label))
            (done (gen-label))
-           (dx-p (node-stack-allocate-p node))
-           (alloc-area-tn (if dx-p csp-tn alloc-tn)))
+           (leave-pa (gen-label))
+           (dx-p (node-stack-allocate-p node)))
       (move context context-arg)
       (move count count-arg)
       ;; Check to see if there are any arguments.
       (inst beq count done)
-      (move result null-tn t)
+      (inst move result null-tn)
 
       ;; We need to do this atomically.
-      (pseudo-atomic (pa-flag)
-        (when dx-p
-          (align-csp temp))
-        ;; Allocate a cons (2 words) for each item.
-        (inst srl result alloc-area-tn n-lowtag-bits)
-        (inst sll result n-lowtag-bits)
-        (inst or result list-pointer-lowtag)
+      (pseudo-atomic (pa-flag :elide-if dx-p)
+        (inst sll count 1)
+        (cond (dx-p
+               (allocation 'list count result list-pointer-lowtag `(,pa-flag ,temp)
+                           :stackp t))
+              (t
+               (let ((end-addr pa-flag)
+                     (new-freeptr temp)
+                     (continue (gen-label)))
+                 (without-scheduling ()
+                   (inst lw result null-tn (- cons-region nil-value))
+                   (inst lw end-addr null-tn (+ 4 (- cons-region nil-value)))
+                   (inst add new-freeptr result count)
+                   ;; Use a different 'code' field for listify. Technically there are enough
+                   ;; bits in the code field to indicate how many bytes to skip to branch out
+                   ;; of the entire loop, which the C handler could use to adjust the RA.
+                   ;; But the assembler doesn't provide a way to compute that distance.
+                   (inst tltu end-addr new-freeptr (logior #x300 (reg-tn-encoding result)))
+                   ;; Jump to the freeptr writeback
+                   (inst beq zero-tn continue)
+                   ;; Encode CONTEXT into a dummy instruction
+                   (inst addu zero-tn context context) ; ADDU never signals overflow
+                   ;; If we return here, then skip the loop
+                   (inst beq zero-tn leave-pa)
+                   (inst nop)
+                   (emit-label continue)
+                   (inst sw new-freeptr null-tn (- cons-region nil-value))
+                   (inst or result list-pointer-lowtag)))))
+
         (move dst result)
-        (inst sll temp count 1)
+
         (inst b enter)
-        (inst addu alloc-area-tn temp)
+        (inst srl count 1)
 
         ;; Store the current cons in the cdr of the previous cons.
         (emit-label loop)
@@ -1161,7 +1222,8 @@ default-value-8
         (storew temp dst 0 list-pointer-lowtag)
 
         ;; NIL out the last cons.
-        (storew null-tn dst 1 list-pointer-lowtag))
+        (storew null-tn dst 1 list-pointer-lowtag)
+        (emit-label leave-pa))
       (emit-label done))))
 
 ;;; Return the location and size of the more arg glob created by Copy-More-Arg.
@@ -1174,9 +1236,9 @@ default-value-8
 ;;; supplied - fixed, and return a pointer that many words below the current
 ;;; stack top.
 ;;;
-(define-vop (more-arg-context)
+(define-vop ()
   (:policy :fast-safe)
-  (:translate sb!c::%more-arg-context)
+  (:translate sb-c::%more-arg-context)
   (:args (supplied :scs (any-reg)))
   (:arg-types tagged-num (:constant fixnum))
   (:info fixed)
@@ -1197,30 +1259,36 @@ default-value-8
   (:vop-var vop)
   (:save-p :compute-only)
   (:generator 3
-    (let ((err-lab
-            (generate-error-code vop 'invalid-arg-count-error nargs)))
-      (cond ((not min)
-             (cond ((zerop max)
-                    (inst bne nargs err-lab))
-                   (t
-                    (inst li temp (fixnumize max))
-                    (inst bne nargs temp err-lab)))
-             (inst nop))
-            (max
-             (when (plusp min)
-               (inst li temp (fixnumize min))
-               (inst sltu temp nargs temp)
-               (inst bne temp err-lab)
-               (inst nop))
-             (inst li temp (fixnumize max))
-             (inst sltu temp temp nargs)
-             (inst bne temp err-lab)
-             (inst nop))
-            ((plusp min)
-             (inst li temp (fixnumize min))
-             (inst sltu temp nargs temp)
-             (inst bne temp err-lab)
-             (inst nop))))))
+   (cond
+     ((not min) ; fixed args
+      (unless (zerop max) (inst li temp (fixnumize max)))
+      (note-this-location vop :internal-error)
+      (inst tne nargs (if (zerop max) zero-tn temp) invalid-arg-count-trap))
+     (t
+      (let ((err-lab
+             (generate-error-code vop 'invalid-arg-count-error nargs)))
+       (cond ((not min)
+              (cond ((zerop max)
+                     (inst bne nargs err-lab))
+                    (t
+                     (inst li temp (fixnumize max))
+                     (inst bne nargs temp err-lab)))
+              (inst nop))
+             (max
+              (when (plusp min)
+                (inst li temp (fixnumize min))
+                (inst sltu temp nargs temp)
+                (inst bne temp err-lab)
+                (inst nop))
+              (inst li temp (fixnumize max))
+              (inst sltu temp temp nargs)
+              (inst bne temp err-lab)
+              (inst nop))
+             ((plusp min)
+              (inst li temp (fixnumize min))
+              (inst sltu temp nargs temp)
+              (inst bne temp err-lab)
+              (inst nop))))))))
 
 ;;; Single-stepping
 
@@ -1229,13 +1297,13 @@ default-value-8
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 3
-    (load-symbol-value stepping sb!impl::*stepping*)
+    (load-symbol-value stepping sb-impl::*stepping*)
     ;; If it's not 0, trap.
     (inst beq stepping DONE)
     (inst nop)
     ;; CONTEXT-PC will be pointing here when the interrupt is handled,
     ;; not after the BREAK.
-    (note-this-location vop :step-before-vop)
+    (note-this-location vop :internal-error)
     ;; CALLEE-REGISTER-OFFSET isn't needed for before-traps, so we
     ;; can just use a bare SINGLE-STEP-BEFORE-TRAP as the code.
     (inst break 0 single-step-before-trap)
