@@ -11,14 +11,12 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!REGALLOC")
+(in-package "SB-REGALLOC")
 
 ;;; for debugging: some parameters controlling which optimizations we
 ;;; attempt
 (defvar *pack-assign-costs* t)
 (defvar *pack-optimize-saves* t)
-;;; FIXME: Perhaps SB-FLUID should be renamed to SB-TWEAK and these
-;;; should be made conditional on SB-TWEAK.
 
 (declaim (ftype (function (component) index) ir2-block-count))
 
@@ -323,7 +321,7 @@
   (let* ((vop (tn-ref-vop op))
          (args (vop-args vop))
          (results (vop-results vop))
-         (name (with-simple-output-to-string (stream)
+         (name (%with-output-to-string (stream)
                  (print-tn-guts tn stream)))
          (2comp (component-info *component-being-compiled*))
          temp)
@@ -340,9 +338,7 @@
      ((setq temp (position-in #'tn-ref-across tn (vop-temps vop)
                               :key #'tn-ref-tn))
       `("~2D: ~A (temporary ~A)" ,loc ,name
-        ,(operand-parse-name (elt (vop-parse-temps
-                                   (vop-parse-or-lose
-                                    (vop-info-name  (vop-info vop))))
+        ,(operand-parse-name (elt (vop-parse-temps (vop-parse-or-lose (vop-name vop)))
                                   temp))))
      ((eq (tn-kind tn) :component)
       `("~2D: ~A (component live)" ,loc ,name))
@@ -353,6 +349,61 @@
      (t
       `("~2D: not referenced?" ,loc)))))
 
+(eval-when (:compile-toplevel)
+  (assert (zerop (rem sb-vm:finite-sc-offset-limit 8)))) ; multiple of 8
+
+(defmacro do-sc-locations ((location locations &optional result (increment 1))
+                           &body body)
+  (let ((inc '#:increment)
+        (bitmap '#:bits)
+        (bias '#:bias))
+    (declare (ignorable inc))
+    `(let ((,bitmap ,locations)
+           (,inc (truly-the (member 1 2 4) ,increment))
+           (,bias 0))
+       ;; This bitmap won't be a machine word if #+sparc but who cares?
+       (declare (type sb-vm:finite-sc-offset-map ,bitmap))
+       ;; BIAS can take on the limiting value at its final iteration
+       (declare (type (integer 0 ,sb-vm:finite-sc-offset-limit) ,bias))
+       (block nil
+         #|
+         ;; This should be faster since there is no loop nesting
+         ;; and only the 1 bits are visited. But it didn't benchmark better.
+         ;; Must dig deeper.
+         #+(and (or x86 x86-64) (not sb-xc-host))
+         (unless (zerop ,bitmap)
+           (loop named #:noname
+                 do
+             (let ((.index. (truly-the
+                             (integer 0 (,sb-vm:n-word-bits))
+                             (%primitive sb-vm::unsigned-word-find-first-bit ,bitmap))))
+               ;; BIAS is the bit index in the original LOCATIONS map
+               ;; which is at index 0 in BITMAP
+               (let ((,location (truly-the sb-vm:finite-sc-offset (+ .index. ,bias))))
+                 ,@body
+                 ;; Shift using 2 steps due to the type constraint on digit-logical-shift-right
+                 ;; and/or CPU behavior of oversized shift.
+                 ;; Consider: 32 bit word and bit 31 (and only that) is on. After processing it,
+                 ;; we need to shift out 32 bits to get to the "next" bit. There is no next bit,
+                 ;; so the word should become 0. But shifting by 32 = shifting by 0 which won't do
+                 ;; the right thing.
+                 (setq ,bitmap (sb-bignum:%digit-logical-shift-right
+                                 (sb-bignum:%digit-logical-shift-right ,bitmap .index.)
+                                 ,inc))
+                 (when (zerop ,bitmap) (loop-finish))
+                 (incf ,bias (truly-the (integer 0 ,sb-vm:n-word-bits)
+                                        (+ .index. ,inc)))))))
+         #-(and (or x86 x86-64) (not sb-xc-host))
+         |#
+         (loop named #:outer repeat (/ sb-vm:finite-sc-offset-limit 8)
+               do (when (ldb-test (byte 8 ,bias) ,bitmap)
+                    ;; scan 8 bits starting at BIAS
+                    (loop named #:inner
+                          for ,location from ,bias to (+ ,bias 7) by ,inc
+                          when (logbitp ,location ,bitmap) do (progn ,@body)))
+                  (incf ,bias 8))
+         ,result))))
+
 ;;; If load TN packing fails, try to give a helpful error message. We
 ;;; find a TN in each location that conflicts, and print it.
 (defun failed-to-pack-load-tn-error (scs op)
@@ -391,7 +442,7 @@
                 time. Recompile.~%Compilation order may be incorrect.~]"
                (mapcar #'sc-name scs)
                n arg-p
-               (vop-info-name (vop-info (tn-ref-vop op)))
+               (vop-name (tn-ref-vop op))
                (unused) (used)
                incon))))
 
@@ -420,6 +471,9 @@
 
 ;;;; register saving
 
+(declaim (start-block optimized-emit-saves emit-saves assign-tn-costs
+                      pack-save-tn))
+
 ;;; Do stuff to note that TN is spilled at VOP for the debugger's benefit.
 (defun note-spilled-tn (tn vop)
   (when (and (tn-leaf tn) (vop-save-set vop))
@@ -467,7 +521,7 @@
     (aver (eq (ir2-block-block block) (ir2-block-block (vop-block vop))))
     (do ((current last (vop-prev current)))
         ((null current))
-      (when (eq (vop-info-name (vop-info current)) name)
+      (when (eq (vop-name current) name)
         (return-from reverse-find-vop current)))))
 
 ;;; For TNs that have other than one writer, we save the TN before
@@ -486,11 +540,14 @@
     (when (eq (tn-kind save) :specified-save)
       (setf (tn-kind save) :save))
     (aver (eq (tn-kind save) :save))
-    (emit-operand-load node block tn save
-                       (if (eq (vop-info-move-args (vop-info vop))
-                               :local-call)
-                           (reverse-find-vop 'allocate-frame vop)
-                           vop))
+    (multiple-value-bind (before block)
+        (if (eq (vop-info-move-args (vop-info vop)) :local-call)
+            (let ((before (reverse-find-vop 'allocate-frame vop)))
+              ;; Because of SPLIT-IR2-BLOCKS the ALLOCATE-FRAME VOP
+              ;; may be in a different block.
+              (values before (vop-block before)))
+            (values vop block))
+      (emit-operand-load node block tn save before))
     (emit-operand-load node block save tn next)))
 
 ;;; Return a VOP after which is an OK place to save the value of TN.
@@ -518,8 +575,7 @@
                       (return nil))))
          (tn-ref-vop res)))
 
-    (unless (eq (vop-info-name (vop-info (tn-ref-vop write)))
-                'move-operand)
+    (unless (eq (vop-name (tn-ref-vop write)) 'move-operand)
       (when res (return nil))
       (setq res write))))
 
@@ -550,14 +606,26 @@
 ;;; appropriate. This is also called by SPILL-AND-PACK-LOAD-TN.
 (defun basic-save-tn (tn vop)
   (declare (type tn tn) (type vop vop))
-  (let ((save (tn-save-tn tn)))
-    (cond ((and save (eq (tn-kind save) :save-once))
-           (restore-single-writer-tn tn vop))
-          ((save-single-writer-tn tn)
-           (restore-single-writer-tn tn vop))
-          (t
-           (save-complex-writer-tn tn vop))))
-  (values))
+  (let* ((save (tn-save-tn tn))
+         (node (vop-node vop)))
+    (flet ((restore ()
+             (not
+              (and (sb-c::combination-p node)
+                   ;; Don't restore if the function doesn't return.
+                   (let ((type (sb-c::lvar-fun-type (sb-c::combination-fun node))))
+                     (and (fun-type-p type)
+                          (eq (sb-kernel:fun-type-returns type) *empty-type*)))
+                   (or
+                    (sb-c::combination-fun-info node)
+                    (policy node (zerop safety)))))))
+      (cond ((and save (eq (tn-kind save) :save-once))
+             (when (restore)
+               (restore-single-writer-tn tn vop)))
+            ((save-single-writer-tn tn)
+             (when (restore)
+               (restore-single-writer-tn tn vop)))
+            (t
+             (save-complex-writer-tn tn vop))))))
 
 ;;; Scan over the VOPs in BLOCK, emiting saving code for TNs noted in
 ;;; the codegen info that are packed into saved SCs.
@@ -658,7 +726,7 @@
       (do ((vop (ir2-block-last-vop block) (vop-prev vop)))
           ((null vop))
         (let ((info (vop-info vop)))
-          (case (vop-info-name info)
+          (case (vop-name vop)
             (allocate-frame
              (aver skipping)
              (setq skipping nil))
@@ -777,7 +845,7 @@
           (when (eq (vop-info-save-p (vop-info vop)) t)
             (let ((penalty (+ save-penalty (vop-depth-cost vop))))
               (do-live-tns (tn (vop-save-set vop) block)
-                #!-fp-and-pc-standard-save
+                #-fp-and-pc-standard-save
                 (let ((save-tn (tn-save-tn tn)))
                   (when (and save-tn (eq :specified-save (tn-kind save-tn)))
                     ;; If we're expecting to spill a TN with a
@@ -808,7 +876,7 @@
 ;;; through and force TNs with specified save locations (OCFP and LRA
 ;;; save locations) to the stack if they are going to be spilled.  See
 ;;; the comment in ASSIGN-TN-COSTS for consequences of not doing so.
-#!-fp-and-pc-standard-save
+#-fp-and-pc-standard-save
 (defun maybe-force-specified-saves-to-stack (component)
   (do-ir2-blocks (block component)
     (do ((vop (ir2-block-start-vop block) (vop-next vop)))
@@ -825,9 +893,23 @@
               ;; race conditions in the debugger involving
               ;; backtraces from asynchronous interrupts.
               (setf (tn-sc tn) (tn-sc save-tn)))))))))
+
+(declaim (end-block))
+
+;; Misc. utilities
+(declaim (maybe-inline unbounded-sc-p))
+(defun unbounded-sc-p (sc)
+  (eq (sb-kind (sc-sb sc)) :unbounded))
+
+(defun unbounded-tn-p (tn)
+  #-sb-xc-host (declare (inline unbounded-sc-p))
+  (unbounded-sc-p (tn-sc tn)))
+
 
 ;;;; load TN packing
 
+(declaim (start-block pack-load-tns load-tn-conflicts-in-sc))
+
 ;;; These variables indicate the last location at which we computed
 ;;; the Live-TNs. They hold the BLOCK and VOP values that were passed
 ;;; to COMPUTE-LIVE-TNS.
@@ -955,7 +1037,7 @@
 ;;;
 ;;; We return a conflicting TN if there is a conflict.
 (defun load-tn-offset-conflicts-in-sb (op sb offset)
-  (declare (type tn-ref op) (type sb!c::finite-sb-template sb) (type index offset))
+  (declare (type tn-ref op) (type sb-c::finite-sb-template sb) (type index offset))
   (aver (eq (sb-kind sb) :finite))
   (let ((vop (tn-ref-vop op)))
     (labels ((tn-overlaps (tn)
@@ -1031,7 +1113,7 @@
       (let* ((tn (tn-ref-tn target))
              (loc (tn-offset tn)))
         (if (and (eq (tn-sc tn) sc)
-                 (sc-locations-member (the sb!vm:finite-sc-offset loc)
+                 (sc-locations-member (the sb-vm:finite-sc-offset loc)
                                       (sc-locations sc))
                  (not (load-tn-conflicts-in-sc op sc loc nil)))
             loc
@@ -1049,7 +1131,7 @@
       (let* ((tn (tn-ref-tn target))
              (loc (tn-offset tn)))
         (when (and (eq (sc-sb sc) (sc-sb (tn-sc tn)))
-                   (sc-locations-member (the sb!vm:finite-sc-offset loc)
+                   (sc-locations-member (the sb-vm:finite-sc-offset loc)
                                         (sc-locations sc))
                    (not (load-tn-conflicts-in-sc op sc loc nil)))
               (return-from select-load-tn-location loc)))))
@@ -1074,7 +1156,7 @@
              (do ((ref refs (tn-ref-next ref)))
                  ((null ref))
                (let ((vop (tn-ref-vop ref)))
-                 (if (eq (vop-info-name (vop-info vop)) 'move-operand)
+                 (if (eq (vop-name vop) 'move-operand)
                      (delete-vop vop)
                      (pushnew (vop-block vop) *repack-blocks*))))))
       (zot (tn-reads tn))
@@ -1184,7 +1266,7 @@
 ;;; the restriction, we pack a Load-TN and load the operand into it.
 ;;; If a load-tn has already been allocated, we can assume that the
 ;;; restriction is satisfied.
-#!-sb-fluid (declaim (inline check-operand-restrictions))
+(declaim (inline check-operand-restrictions))
 (defun check-operand-restrictions (scs ops)
   (declare (list scs) (type (or tn-ref null) ops))
 
@@ -1195,7 +1277,8 @@
     (let ((target (tn-ref-target op))
           (tn (tn-ref-tn op)))
       (when (and target
-                 (not (eq (tn-kind tn) :unused)))
+                 (not (eq (tn-kind tn) :unused))
+                 (tn-primitive-type tn))
         (let* ((load-tn (tn-ref-load-tn op))
                (load-scs (svref (car scs)
                                 (sc-number
@@ -1212,7 +1295,8 @@
     (let ((target (tn-ref-target op))
           (tn (tn-ref-tn op)))
       (unless (or target
-                  (eq (tn-kind tn) :unused))
+                  (eq (tn-kind tn) :unused)
+                  (not (tn-primitive-type tn)))
         (let* ((load-tn (tn-ref-load-tn op))
                (load-scs (svref (car scs)
                                 (sc-number
@@ -1243,6 +1327,10 @@
 
 ;;;; targeting
 
+(declaim (start-block pack pack-tn target-if-desirable
+                      ;; needed for pack-iterative
+                      pack-wired-tn))
+
 ;;; Link the TN-REFS READ and WRITE together using the TN-REF-TARGET
 ;;; when this seems like a good idea. Currently we always do, as this
 ;;; increases the success of load-TN targeting.
@@ -1321,7 +1409,7 @@
                           &rest keys &key limit reads writes)
                          &body body)
   (declare (ignore limit reads writes))
-  (let ((callback (sb!xc:gensym "CALLBACK")))
+  (let ((callback (gensym "CALLBACK")))
     `(flet ((,callback (,target-variable)
               ,@body))
        (declare (dynamic-extent #',callback))
@@ -1385,15 +1473,6 @@
 
 ;;;; pack interface
 
-;; Misc. utilities
-(declaim (inline unbounded-sc-p))
-(defun unbounded-sc-p (sc)
-  (eq (sb-kind (sc-sb sc)) :unbounded))
-
-(defun unbounded-tn-p (tn)
-  (unbounded-sc-p (tn-sc tn)))
-(declaim (notinline unbounded-sc-p))
-
 ;;; Attempt to pack TN in all possible SCs, first in the SC chosen by
 ;;; representation selection, then in the alternate SCs in the order
 ;;; they were specified in the SC definition. If the TN-COST is
@@ -1467,46 +1546,32 @@
     ;; For non-x86 ports the presence of a save-tn associated with a
     ;; tn is used to identify the old-fp and return-pc tns. It depends
     ;; on the old-fp and return-pc being passed in registers.
-    #!-fp-and-pc-standard-save
-    (when (and (not (eq (tn-kind tn) :specified-save))
+    (when (and #-fp-and-pc-standard-save
+               (not (eq (tn-kind tn) :specified-save))
                (conflicts-in-sc original sc offset))
-      (error "~S is wired to a location that it conflicts with." tn))
+      (error "~S is wired to location ~D in SC ~A of kind ~S that it conflicts with."
+             tn offset sc (tn-kind tn)))
 
     ;; Use the above check, but only print a verbose warning. This can
     ;; be helpful for debugging the x86 port.
     #+nil
     (when (and (not (eq (tn-kind tn) :specified-save))
                (conflicts-in-sc original sc offset))
-          (format t "~&* Pack-wired-tn possible conflict:~%  ~
+      (format t "~&* Pack-wired-tn possible conflict:~%  ~
                      tn: ~S; tn-kind: ~S~%  ~
                      sc: ~S~%  ~
                      sb: ~S; sb-name: ~S; sb-kind: ~S~%  ~
                      offset: ~S; end: ~S~%  ~
                      original ~S~%  ~
                      tn-save-tn: ~S; tn-kind of tn-save-tn: ~S~%"
-                  tn (tn-kind tn) sc
-                  sb (sb-name sb) (sb-kind sb)
-                  offset end
-                  original
-                  (tn-save-tn tn) (tn-kind (tn-save-tn tn))))
-
-    ;; On the x86 ports the old-fp and return-pc are often passed on
-    ;; the stack so the above hack for the other ports does not always
-    ;; work. Here the old-fp and return-pc tns are identified by being
-    ;; on the stack in their standard save locations.
-    #!+fp-and-pc-standard-save
-    (when (and (not (and
-                     (= (sc-number sc) #.(sc+offset-scn old-fp-passing-offset))
-                     (= offset #.(sc+offset-offset old-fp-passing-offset))))
-               (not (and
-                     (= (sc-number sc) #.(sc+offset-scn return-pc-passing-offset))
-                     (= offset #.(sc+offset-offset return-pc-passing-offset))))
-               (conflicts-in-sc original sc offset))
-      (error "~S is wired to location ~D in SC ~A of kind ~S that it conflicts with."
-             tn offset sc (tn-kind tn)))
+              tn (tn-kind tn) sc
+              sb (sb-name sb) (sb-kind sb)
+              offset end
+              original
+              (tn-save-tn tn) (tn-kind (tn-save-tn tn))))
 
     (unless (eq (sb-kind sb) :unbounded)
-      (setf (ldb (byte 1 (truly-the sb!vm:finite-sc-offset offset))
+      (setf (ldb (byte 1 (truly-the sb-vm:finite-sc-offset offset))
                  (finite-sb-wired-map sb))
             1))
     (add-location-conflicts original sc offset)))
@@ -1613,7 +1678,7 @@
          ;; be packed on the stack, and which are important not to spill.
          (if *pack-assign-costs*
              (assign-tn-costs component)
-             #!-fp-and-pc-standard-save
+             #-fp-and-pc-standard-save
              (maybe-force-specified-saves-to-stack component))
 
          ;; Actually allocate registers for most TNs. After this, only
