@@ -15,6 +15,85 @@
 
 ;;;; CL:COMPILE
 
+(defun ir1-toplevel-for-compile (form name path)
+  (let* ((*current-path* path)
+         (component (make-empty-component))
+         (*current-component* component)
+         (debug-name-tail (or name (name-lambdalike form)))
+         (source-name (or name '.anonymous.)))
+    (setf (component-name component) (debug-name 'initial-component debug-name-tail)
+          (component-kind component) :initial)
+    (let* ((fun (let ((*allow-instrumenting* t))
+                  (ir1-convert-lambdalike form
+                                          :source-name source-name)))
+           ;; Convert the XEP using the policy of the real function. Otherwise
+           ;; the wrong policy will be used for deciding whether to type-check
+           ;; the parameters of the real function (via CONVERT-CALL /
+           ;; PROPAGATE-TO-ARGS). -- JES, 2007-02-27
+           (*lexenv* (make-lexenv :policy (lexenv-policy (functional-lexenv fun))))
+           (xep (ir1-convert-lambda (make-xep-lambda-expression fun)
+                                    :source-name source-name
+                                    :debug-name (debug-name 'tl-xep debug-name-tail))))
+      (when name
+        (assert-global-function-definition-type name fun))
+      (setf (functional-kind xep) :external
+            (functional-entry-fun xep) fun
+            (functional-entry-fun fun) xep
+            (component-reanalyze component) t
+            (functional-has-external-references-p xep) t)
+      (reoptimize-component component :maybe)
+      (locall-analyze-xep-entry-point fun)
+      ;; Any leftover REFs to FUN outside local calls get replaced with the
+      ;; XEP.
+      (substitute-leaf-if (lambda (ref)
+                            (let* ((lvar (ref-lvar ref))
+                                   (dest (when lvar (lvar-dest lvar)))
+                                   (kind (when (basic-combination-p dest)
+                                           (basic-combination-kind dest))))
+                              (neq :local kind)))
+                          xep
+                          fun)
+      xep)))
+
+;;; Compile LAMBDA-EXPRESSION and return the compiled FUNCTION value.
+;;;
+;;; If NAME is provided, then we try to use it as the name of the
+;;; function for debugging/diagnostic information.
+(defun %compile (form ephemeral name tlf)
+  (when name
+    (legal-fun-name-or-type-error name))
+  (with-ir1-namespace
+    (let* ((*lexenv* (make-lexenv
+                      :policy *policy*
+                      :handled-conditions *handled-conditions*
+                      :disabled-package-locks *disabled-package-locks*))
+           (*compile-object* (make-core-object ephemeral))
+           (path `(original-source-start 0 ,tlf))
+           (lambda (ir1-toplevel-for-compile form name path)))
+
+      ;; FIXME: The compile-it code from here on is sort of a
+      ;; twisted version of the code in COMPILE-TOPLEVEL. It'd be
+      ;; better to find a way to share the code there; or
+      ;; alternatively, to use this code to replace the code there.
+      ;; (The second alternative might be pretty easy if we used
+      ;; the :LOCALL-ONLY option to IR1-FOR-LAMBDA. Then maybe the
+      ;; whole FUNCTIONAL-KIND=:TOPLEVEL case could go away..)
+
+      (locall-analyze-clambdas-until-done (list lambda))
+
+      (dolist (component (find-initial-dfo (list lambda)))
+        (compile-component component))
+
+      (let ((object *compile-object*))
+        (multiple-value-bind (res found-p)
+            (gethash (leaf-info lambda) (core-object-entry-table object))
+          (aver found-p)
+          (fix-core-source-info *source-info* object
+                                (and (policy (lambda-bind lambda)
+                                         (> store-source-form 0))
+                                     res))
+          res)))))
+
 ;;; Handle the following:
 ;;;  - CL:COMPILE when the argument is not already a compiled function.
 ;;;  - %SIMPLE-EVAL in "pretend we don't have an interpreter" mode
@@ -26,7 +105,7 @@
 (defun compile-in-lexenv (form *lexenv* name source-info tlf ephemeral errorp)
   (let ((source-paths (when source-info *source-paths*)))
     (with-compilation-values
-      (sb-xc:with-compilation-unit ()
+      (with-compilation-unit ()
         ;; FIXME: These bindings were copied from SUB-COMPILE-FILE with
         ;; few changes. Once things are stable, the shared bindings
         ;; probably be merged back together into some shared utility
@@ -42,13 +121,13 @@
                   (*source-info* (or source-info
                                   (make-lisp-source-info
                                    form :parent *source-info*)))
-                  (*toplevel-lambdas* ())
-                  (*block-compile* nil)
                   (*allow-instrumenting* nil)
-                  (*compiler-coverage-metadata* nil)
-                  (*msan-unpoison*
-                   (and (member :msan *features*)
-                        (find-dynamic-foreign-symbol-address "__msan_unpoison")))
+                  (*compilation*
+                   (make-compilation
+                    :msan-unpoison
+                    (and (member :msan *features*)
+                         (find-dynamic-foreign-symbol-address "__msan_unpoison"))
+                    :block-compile nil))
                   (*current-path* nil)
                   (*last-message-count* (list* 0 nil nil))
                   (*last-error-context* nil)
@@ -74,6 +153,13 @@
                   ;; controlled by function arguments and lexical variables.
                   (*compile-verbose* nil)
                   (*compile-print* nil)
+                  ;; in some circumstances, we can trigger execution
+                  ;; of user code during optimization, which can
+                  ;; re-enter the compiler through explicit calls to
+                  ;; EVAL or COMPILE.  Those inner evaluations
+                  ;; shouldn't attempt to report any compiler problems
+                  ;; using the outer compiler error context.
+                  (*compiler-error-context* nil)
                   (oops nil))
              (handler-bind (((satisfies handle-condition-p) #'handle-condition-handler))
                (unless source-paths
@@ -85,9 +171,7 @@
                          ;; the error came from, not how the compiler got there.
                          (go :error))))
                  (return
-                     (%compile form (make-core-object ephemeral)
-                               :name name
-                               :path `(original-source-start 0 ,tlf)))))
+                   (%compile form ephemeral name tlf))))
            :error
              ;; Either signal the error right away, or return a function that
              ;; will signal the corresponding COMPILED-PROGRAM-ERROR. This is so
@@ -108,6 +192,22 @@
                               :message message
                               :source source)))))))))))
 
+;;; NOTE: COMPILE may be slightly nonconforming regarding generic functions,
+;;; but no more nonconforming than it was prior to the redefinition of
+;;; COMPILED-FUNCTION to exclude GENERIC-FUNCTION.
+;;; The concern stems from http://www.lispworks.com/documentation/HyperSpec/Issues/iss064_w.htm
+;;; which says "(4) Clarify that COMPILE must produce an object of type COMPILED-FUNCTION."
+;;;
+;;; In the case where DEFINITION is given, we're fine: the compiler can only return
+;;; a compiled function. But if only NAME is given, and it is fboundp to a generic-function,
+;;; we don't do anything at all - we don't touch the GF's dispatch function (which is a closure
+;;; over compiled code) and we don't touch the methods. But COMPILE doesn't return a function
+;;; in that case, so it's not wrong that COMPILED-FUNCTION-P is false of the result,
+;;; because the result is a symbol, not a function.
+
+;;; Also note that we lack good regression tests setting expectations around what's supposed
+;;; to happen when DEFINITION is supplied as a generic function. (Does it even make sense?)
+
 (defun compile (name &optional (definition (or (and (symbolp name)
                                                     (macro-function name))
                                                (fdefinition name))))
@@ -129,24 +229,33 @@ WARNING occur during the compilation, and NIL otherwise.
 Tertiary value is true if any conditions of type ERROR, or WARNING that are
 not STYLE-WARNINGs occur during compilation, and NIL otherwise.
 "
-  (multiple-value-bind (compiled-definition warnings-p failure-p)
-      ;; TODO: generic functions with any interpreted methods
-      ;; should compile the methods and reinstall them.
-      (if (compiled-function-p definition)
-          (values definition nil nil)
+  (binding*
+     ((clock-start (get-thread-virtual-time))
+      ((compiled-definition warnings-p failure-p)
+      (if (or (compiled-function-p definition)
+              (sb-pcl::generic-function-p definition))
+          ;; We're not invoking COMPILE. This is a minor bug if this is
+          ;; a GENERIC-FUNCTION whose methods are interpreted.
+          (values (make-unbound-marker) nil nil)
           (multiple-value-bind (sexpr lexenv)
               (if (not (typep definition 'interpreted-function))
                   (values (the cons definition) (make-null-lexenv))
                   #+(or sb-eval sb-fasteval)
                   (prepare-for-compile definition))
-            (compile-in-lexenv sexpr lexenv name nil nil nil nil)))
+            (sb-vm:without-arena "compile"
+              (compile-in-lexenv sexpr lexenv name nil nil nil nil))))))
+    (accumulate-compiler-time '*compile-elapsed-time* clock-start)
     (values (cond (name
-                   (if (and (symbolp name) (macro-function name))
-                       (setf (macro-function name) compiled-definition)
-                       (setf (fdefinition name) compiled-definition))
+                   ;; Do NOT assign anything into the symbol if we did not
+                   ;; actually invoke the compiler
+                   (unless (unbound-marker-p compiled-definition)
+                     (if (and (symbolp name) (macro-function name))
+                         (setf (macro-function name) compiled-definition)
+                         (setf (fdefinition name) compiled-definition)))
                    name)
-                  (t
-                   compiled-definition))
+                  ;; Didn't run the compiler
+                  ((unbound-marker-p compiled-definition) definition)
+                  (t compiled-definition))
             warnings-p
             failure-p)))
 
@@ -272,7 +381,7 @@ not STYLE-WARNINGs occur during compilation, and NIL otherwise.
 ;; answers. (Modulo any bugs due to near-total lack of testing)
 
 (defun compile-file-position-helper (file-info path-to-find)
-  (let (found-form start-char)
+  (let (start-char)
     (labels
         ((recurse (subpath upper-bound queue)
            (let ((index -1))
@@ -288,8 +397,7 @@ not STYLE-WARNINGs occur during compilation, and NIL otherwise.
                   ;; This does not eagerly declare victory, because we want
                   ;; to find the rightmost match. In "#1=(FOO)" there are two
                   ;; different annotations pointing to (FOO).
-                  (setq found-form (cdr item)
-                        start-char (caar item)))
+                  (setq start-char (caar item)))
                 (unless queue (return))
                 (let* ((next (car queue))
                        (next-end (cdar next)))
@@ -320,9 +428,6 @@ not STYLE-WARNINGs occur during compilation, and NIL otherwise.
 
 ;;;; Coverage helpers
 
-(defun record-code-coverage (namestring cc)
-  (setf (gethash namestring (car *code-coverage-info*)) cc))
-
 (defun clear-code-coverage ()
   (clrhash (car *code-coverage-info*))
   (setf (cdr *code-coverage-info*) nil))
