@@ -14,29 +14,9 @@
 (defconstant arg-count-sc (make-sc+offset any-reg-sc-number rcx-offset))
 (defconstant closure-sc (make-sc+offset any-reg-sc-number rax-offset))
 
-;;; Make a passing location TN for a local call return PC.
-;;;
-;;; Always wire the return PC location to the stack in its standard
-;;; location.
-(defun make-return-pc-passing-location (standard)
-  (declare (ignore standard))
-  (make-wired-tn (primitive-type-or-lose 'system-area-pointer)
-                 sap-stack-sc-number return-pc-save-offset))
-
 (defconstant return-pc-passing-offset
   (make-sc+offset sap-stack-sc-number return-pc-save-offset))
 
-;;; This is similar to MAKE-RETURN-PC-PASSING-LOCATION, but makes a
-;;; location to pass OLD-FP in.
-;;;
-;;; This is wired in both the standard and the local-call conventions,
-;;; because we want to be able to assume it's always there. Besides,
-;;; the x86 doesn't have enough registers to really make it profitable
-;;; to pass it in a register.
-(defun make-old-fp-passing-location ()
-  (make-wired-tn *fixnum-primitive-type* control-stack-sc-number
-                 ocfp-save-offset))
-
 (defconstant old-fp-passing-offset
   (make-sc+offset control-stack-sc-number ocfp-save-offset))
 
@@ -46,16 +26,17 @@
 ;;;
 ;;; Without using a save-tn - which does not make much sense if it is
 ;;; wired to the stack?
-(defun make-old-fp-save-location (physenv)
-  (physenv-debug-live-tn (make-wired-tn *fixnum-primitive-type*
-                                        control-stack-sc-number
-                                        ocfp-save-offset)
-                         physenv))
-(defun make-return-pc-save-location (physenv)
-  (physenv-debug-live-tn
-   (make-wired-tn (primitive-type-or-lose 'system-area-pointer)
-                  sap-stack-sc-number return-pc-save-offset)
-   physenv))
+(defun make-old-fp-save-location ()
+  (let ((tn (make-wired-tn *fixnum-primitive-type*
+                           control-stack-sc-number
+                           ocfp-save-offset)))
+    (setf (tn-kind tn) :environment)
+    tn))
+(defun make-return-pc-save-location ()
+  (let ((tn (make-wired-tn (primitive-type-or-lose 'system-area-pointer)
+                           sap-stack-sc-number return-pc-save-offset)))
+    (setf (tn-kind tn) :environment)
+    tn))
 
 ;;; Make a TN for the standard argument count passing location. We only
 ;;; need to make the standard location, since a count is never passed when we
@@ -258,7 +239,7 @@
 ;;;     there are stack values.
 ;;;  -- Reset SP. This must be done whenever other than 1 value is
 ;;;     returned, regardless of the number of values desired.
-(defun default-unknown-values (vop values nvals node)
+(defun default-unknown-values (vop values nvals node rbx move-temp)
   (declare (type (or tn-ref null) values)
            (type unsigned-byte nvals))
   (let ((type (sb-c::basic-combination-derived-type node)))
@@ -267,17 +248,11 @@
        (note-this-location vop :single-value-return)
        (cond
          ((<= (sb-kernel:values-type-max-value-count type)
-              register-arg-count)
-          (when (and (named-type-p type)
-                     (eq nil (named-type-name type)))
-            ;; The function never returns, it may happen that the code
-            ;; ends right here leavig the :SINGLE-VALUE-RETURN note
-            ;; dangling. Let's emit a NOP.
-            (inst nop)))
+              register-arg-count))
          ((not (sb-kernel:values-type-may-be-single-value-p type))
-          (inst mov rsp-tn rbx-tn))
+          (inst mov rsp-tn rbx))
          (t
-          (inst cmov :c rsp-tn rbx-tn))))
+          (inst cmov :c rsp-tn rbx))))
       ((<= nvals register-arg-count)
        (note-this-location vop :unknown-return)
        (when (sb-kernel:values-type-may-be-single-value-p type)
@@ -298,11 +273,11 @@
                  do
                  (inst mov :dword (tn-ref-tn tn-ref)
                        (if 2nd-tn-live 2nd-tn nil-value)))))
-           (inst mov rbx-tn rsp-tn)
+           (inst mov rbx rsp-tn)
            (emit-label regs-defaulted)))
        (when (< register-arg-count
                 (sb-kernel:values-type-max-value-count type))
-         (inst mov rsp-tn rbx-tn)))
+         (inst mov rsp-tn rbx)))
       (t
        (collect ((defaults))
          (let ((default-stack-slots (gen-label))
@@ -327,7 +302,7 @@
                    (loop for null = nil-value then (car used-registers)
                          for reg in used-registers
                          do (inst mov :dword reg null))
-                   (move rbx-tn rsp-tn)
+                   (move rbx rsp-tn)
                    (inst jmp defaulting-done)))
             REGS-DEFAULTED
             (do ((i register-arg-count (1+ i))
@@ -343,12 +318,12 @@
                     (inst jmp :be default-lab)
                     (sc-case tn
                       (control-stack
-                       (loadw r11-tn rbx-tn (frame-word-offset (+ sp->fp-offset i)))
-                       (inst mov tn r11-tn))
+                       (loadw move-temp rbx (frame-word-offset (+ sp->fp-offset i)))
+                       (inst mov tn move-temp))
                       (t
-                       (loadw tn rbx-tn (frame-word-offset (+ sp->fp-offset i)))))))))
+                       (loadw tn rbx (frame-word-offset (+ sp->fp-offset i)))))))))
             DEFAULTING-DONE
-            (move rsp-tn rbx-tn)
+            (move rsp-tn rbx)
             (let ((defaults (defaults)))
               (when defaults
                 (assemble (:elsewhere)
@@ -356,7 +331,7 @@
                   (loop for null = nil-value then (car used-registers)
                         for reg in used-registers
                         do (inst mov :dword reg null))
-                  (move rbx-tn rsp-tn)
+                  (move rbx rsp-tn)
                   (dolist (default defaults)
                     (emit-label (car default))
                     (inst mov (cdr default) nil-value))
@@ -473,7 +448,7 @@
     (emit-label trampoline-label)
     (popw rbp-tn (frame-word-offset return-pc-save-offset)))
   (when alignp
-    (emit-alignment n-lowtag-bits :long-nop))
+    (emit-alignment n-lowtag-bits alignp))
   (emit-label start-label))
 
 ;;; Non-TR local call for a fixed number of values passed according to
@@ -506,11 +481,12 @@
   (:vop-var vop)
   (:ignore nfp arg-locs args callee)
   (:node-var node)
+  (:temporary (:sc any-reg) move-temp)
   (:generator 5
     (move rbp-tn fp)
     (note-this-location vop :call-site)
     (inst call target)
-    (default-unknown-values vop values nvals node)))
+    (default-unknown-values vop values nvals node rbx-tn move-temp)))
 
 ;;; Non-TR local call for a variable number of return values passed according
 ;;; to the unknown values convention. The results are the start of the values
@@ -613,7 +589,7 @@
 ;;; In tail call with fixed arguments, the passing locations are
 ;;; passed as a more arg, but there is no new-FP, since the arguments
 ;;; have been set up in the current frame.
-(macrolet ((define-full-call (vop-name named return variable)
+(macrolet ((define-full-call (vop-name named return variable &optional args)
             (aver (not (and variable (eq return :tail))))
             #+immobile-code (when named (setq named :direct))
             `(define-vop (,vop-name ,@(when (eq return :unknown)
@@ -634,7 +610,9 @@
                    '((old-fp)
                      (return-pc)))
 
-               ,@(unless variable '((args :more t :scs (descriptor-reg)))))
+               ,@(unless variable
+                   `((args :more t ,@(unless (eq args :fixed)
+                                       '(:scs (descriptor-reg control-stack)))))))
 
                ,@(when (eq return :fixed)
                    '((:results (values :more t))))
@@ -642,7 +620,9 @@
                (:save-p ,(if (eq return :tail) :compute-only t))
 
                ,@(unless (or (eq return :tail) variable)
-               '((:move-args :full-call)))
+                   `((:move-args ,(if (eq args :fixed)
+                                      :fixed
+                                      :full-call))))
 
                (:vop-var vop)
                (:node-var node)
@@ -660,8 +640,7 @@
 
                (:ignore
                ,@(unless (or variable (eq return :tail)) '(arg-locs))
-               ,@(unless variable '(args))
-               ,@(and (eq return :fixed) '(rbx)))
+               ,@(unless variable '(args)))
 
                ;; We pass either the fdefn object (for named call) or
                ;; the actual function object (for unnamed call) in
@@ -682,7 +661,8 @@
                ,@(when (eq return :fixed)
                    ;; Save it for DEFAULT-UNKNOWN-VALUES to work
                    `((:temporary (:sc unsigned-reg :offset rbx-offset
-                                  :from :result) rbx)))
+                                  :from :result) rbx)
+                     (:temporary (:sc any-reg) move-temp)))
 
                ;; With variable call, we have to load the
                ;; register-args out of the (new) stack frame before
@@ -738,9 +718,11 @@
                                for index downfrom -1
                                do (noise `(loadw ,name new-fp ,index)))
                               (noise))
-                   '((if (zerop nargs)
-                         (zeroize rcx)
-                       (inst mov rcx (fixnumize nargs)))))
+                     '((cond ((listp nargs)) ;; no-verify-arg-count
+                             ((zerop nargs)
+                              (zeroize rcx))
+                             (t
+                              (inst mov rcx (fixnumize nargs))))))
                ,@(cond ((eq return :tail)
                         '(;; Python has figured out what frame we should
                           ;; return to so might as well use that clue.
@@ -817,41 +799,21 @@
                DONE
                (note-this-location vop :call-site)
                ,(cond ((eq named :direct)
-                       #+immobile-code
-                       `(let* ((fixup (make-fixup
-                                       fun
-                                       (if (static-fdefn-offset fun)
-                                           :static-call
-                                           :named-call)))
-                               (target
-                                 (if (and (sb-c::code-immobile-p node)
-                                          (not step-instrumenting))
-                                     fixup
-                                     (progn
-                                       ;; RAX-TN was not declared as a temp var,
-                                       ;; however it's sole purpose at this point is
-                                       ;; for function call, so even if it was used
-                                       ;; to compute a stack argument, it's free now.
-                                       ;; If the call hits the undefined fun trap,
-                                       ;; RAX will get loaded regardless.
-                                       (inst mov rax-tn fixup)
-                                       rax-tn))))
-                          (inst ,(if (eq return :tail) 'jmp 'call) target))
-                       #-immobile-code
-                       `(inst ,(if (eq return :tail) 'jmp 'call)
-                              (ea (+ nil-value (static-fun-offset fun)))))
+                       #+immobile-code `(emit-direct-call fun ',(if (eq return :tail) 'jmp 'call)
+                                                          node step-instrumenting)
+                       #-immobile-code `(inst ,(if (eq return :tail) 'jmp 'call)
+                                              (ea (+ nil-value (static-fun-offset fun)))))
                       #-immobile-code
                       (named
                        `(inst ,(if (eq return :tail) 'jmp 'call)
-                              (ea (- (* fdefn-raw-addr-slot n-word-bytes)
-                                     other-pointer-lowtag) rax)))
+                              (object-slot-ea rax fdefn-raw-addr-slot other-pointer-lowtag)))
                       ((eq return :tail)
                        `(tail-call-unnamed rax fun-type vop))
                       (t
                        `(call-unnamed rax fun-type vop)))
                ,@(ecase return
                    (:fixed
-                    '((default-unknown-values vop values nvals node)))
+                    '((default-unknown-values vop values nvals node rbx move-temp)))
                    (:unknown
                     '((note-this-location vop :unknown-return)
                       (receive-unknown-values values-start nvals start count
@@ -872,7 +834,35 @@
   (define-full-call static-tail-call-named :direct :tail nil)
 
   (define-full-call call-variable nil :fixed t)
-  (define-full-call multiple-call-variable nil :unknown t))
+  (define-full-call multiple-call-variable nil :unknown t)
+  (define-full-call fixed-call-named t :fixed nil :fixed)
+  (define-full-call fixed-tail-call-named t :tail nil :fixed))
+
+;;; Call NAME "directly" meaning in a single JMP or CALL instruction,
+;;; if possible (without loading RAX)
+(defun emit-direct-call (name instruction node step-instrumenting)
+      ;; a :STATIC-CALL fixup is the address of the entry point of
+      ;; the function itself, and a :FDEFN-CALL fixup is the address
+      ;; of the JMP instruction embedded in the header for the named FDEFN.
+  (when (static-fdefn-offset name)
+    (let ((fixup (make-fixup name :static-call)))
+      (return-from emit-direct-call
+        (inst* instruction (if (sb-c::code-immobile-p node) fixup (ea fixup))))))
+  (let* ((fixup (make-fixup name :fdefn-call))
+         (target
+              (if (and (sb-c::code-immobile-p node)
+                       (not step-instrumenting))
+                  fixup
+                  (progn
+                    ;; RAX-TN was not declared as a temp var,
+                    ;; however it's sole purpose at this point is
+                    ;; for function call, so even if it was used
+                    ;; to compute a stack argument, it's free now.
+                    ;; If the call hits the undefined fun trap,
+                    ;; RAX will get loaded regardless.
+                    (inst mov rax-tn fixup)
+                    rax-tn))))
+    (inst* instruction target)))
 
 ;;; Invoke the function-designator FUN.
 (defun tail-call-unnamed (fun type vop)
@@ -1272,54 +1262,85 @@
     done))
 
 ;;; Turn more arg (context, count) into a list.
+;;; Cons cells will be filled in right-to-left.
+;;; This has a slight advantage in code size, and eliminates an initial
+;;; forward jump into the loop. it also admits an interesting possibility
+;;; to reduce the scope of the pseudo-atomic section so as not to
+;;; encompass construction of the list. To do that, we will need to invent
+;;; a new widetag for "contiguous CONS block" which has a header conveying
+;;; the total payload length. Initially we would store that into the CAR of the
+;;; first cons cell. Upon seeing such header, GC shall treat that entire object
+;;; as a boxed payload of specified length. It will be implicitly pinned
+;;; (if conservative) or transported as a whole (if precise). Then when the CAR
+;;; of the first cons is overwritten, the object changes to a linked list.
 (define-vop ()
   (:translate %listify-rest-args)
   (:policy :safe)
-  (:args (context :scs (descriptor-reg) :target src)
+  ;; CONTEXT is used throughout the copying loop
+  (:args (context :scs (descriptor-reg) :to :save)
          (count :scs (any-reg) :target rcx))
   (:arg-types * tagged-num)
-  (:temporary (:sc unsigned-reg :offset rsi-offset :from (:argument 0)) src)
+  ;; The only advantage to specifying RCX here is that JRCXZ can be used
+  ;; in one place, and then only in the unlikely scenario that CONTEXT is not
+  ;; in RCX. If it was, SHL sets/clears the Z flag, but LEA doesn't.
+  ;; Not much of an advantage, but why not.
   (:temporary (:sc unsigned-reg :offset rcx-offset :from (:argument 1)) rcx)
-  (:temporary (:sc unsigned-reg :offset rax-offset) rax)
-  (:temporary (:sc unsigned-reg) dst)
+  ;; Note that DST conflicts with RESULT because we use both as temps
+  (:temporary (:sc unsigned-reg) value dst)
+  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
   (:results (result :scs (descriptor-reg)))
   (:node-var node)
   (:generator 20
-    (let ((enter (gen-label))
-          (loop (gen-label))
+    (let ((loop (gen-label))
           (done (gen-label))
+          (leave-pa (gen-label))
           (stack-allocate-p (node-stack-allocate-p node)))
-      (move src context)
-      (move rcx count)
-      ;; Check to see whether there are no args, and just return NIL if so.
+      ;; Compute the number of bytes to allocate
+      (let ((shift (- (1+ word-shift) n-fixnum-tag-bits)))
+        (if (location= count rcx)
+            (inst shl :dword rcx shift)
+            (inst lea :dword rcx (ea nil count (ash 1 shift)))))
+      ;; Setup for the CDR of the last cons (or the entire result) being NIL.
       (inst mov result nil-value)
-      (inst test rcx rcx)
-      (inst jmp :z done)
-      (inst lea dst (ea nil rcx (ash 2 (- word-shift n-fixnum-tag-bits))))
+      (inst jrcxz DONE)
       (unless stack-allocate-p
-        (instrument-alloc dst node))
-      (pseudo-atomic (:elide-if stack-allocate-p)
-       (allocation 'list dst list-pointer-lowtag node stack-allocate-p dst)
-       ;; Set up the result.
-       (move result dst)
-       ;; Jump into the middle of the loop, 'cause that's where we want
-       ;; to start.
-       (inst jmp enter)
-       (emit-label loop)
-       ;; Compute a pointer to the next cons.
-       (inst add dst (* cons-size n-word-bytes))
-       ;; Store a pointer to this cons in the CDR of the previous cons.
-       (storew dst dst -1 list-pointer-lowtag)
-       (emit-label enter)
-       ;; Grab one value and stash it in the car of this cons.
-       (inst mov rax (ea src))
-       (inst sub src n-word-bytes)
-       (storew rax dst 0 list-pointer-lowtag)
-       ;; Go back for more.
-       (inst sub rcx (fixnumize 1))
+        (instrument-alloc +cons-primtype+ rcx node (list value dst) thread-tn))
+      (pseudo-atomic (:elide-if stack-allocate-p :thread-tn thread-tn)
+       ;; Produce an untagged pointer into DST
+       (if stack-allocate-p
+           (stack-allocation rcx 0 dst)
+           (allocation +cons-primtype+ rcx 0 dst node value thread-tn
+                       :overflow
+                       (lambda ()
+                         (inst push rcx)
+                         (inst push context)
+                         (invoke-asm-routine
+                          'call (if (system-tlab-p 0 node) 'sys-listify-&rest 'listify-&rest)
+                          node)
+                         (inst pop result)
+                         (inst jmp leave-pa))))
+       ;; Recalculate DST as a tagged pointer to the last cons
+       (inst lea dst (ea (- list-pointer-lowtag (* cons-size n-word-bytes)) dst rcx))
+       (inst shr :dword rcx (1+ word-shift)) ; convert bytes to number of cells
+       ;; The rightmost arguments are at lower addresses.
+       ;; Start by indexing the last argument
+       (inst neg rcx) ; :QWORD because it's a signed number
+       (emit-label LOOP)
+       ;; Grab one value and store into this cons. Use RCX as an index into the
+       ;; vector of values in CONTEXT, but add 8 because CONTEXT points exactly at
+       ;; the 0th value, which means that the index is 1 word too low.
+       ;; (It's -1 if there is exactly 1 value, instead of 0, and so on)
+       (inst mov value (ea 8 context rcx 8))
+       ;; RESULT began as NIL which gives the correct value for the CDR in the final cons.
+       ;; Subsequently it points to each cons just populated, which is correct all the way
+       ;; up to and including the final result.
+       (storew result dst cons-cdr-slot list-pointer-lowtag)
+       (storew value dst cons-car-slot list-pointer-lowtag)
+       (inst mov result dst) ; preserve the value to put in the CDR of the preceding cons
+       (inst sub dst (* cons-size n-word-bytes)) ; get the preceding cons
+       (inst inc rcx) ; :QWORD because it's a signed number
        (inst jmp :nz loop)
-       ;; NIL out the last cons.
-       (storew nil-value dst 1 list-pointer-lowtag))
+       (emit-label leave-pa))
       (emit-label done))))
 
 ;;; Return the location and size of the &MORE arg glob created by
@@ -1356,6 +1377,7 @@
   (:policy :fast-safe)
   (:args (nargs :scs (any-reg)))
   (:arg-types positive-fixnum (:constant t) (:constant t))
+  (:temporary (:sc unsigned-reg :offset rbx-offset) temp)
   (:info min max)
   (:vop-var vop)
   (:save-p :compute-only)
@@ -1363,24 +1385,24 @@
     ;; NOTE: copy-more-arg expects this to issue a CMP for min > 1
     (let ((err-lab
             (generate-error-code vop 'invalid-arg-count-error nargs)))
-      (flet ((check-min ()
-               (cond ((= min 1)
-                      (inst test nargs nargs)
-                      (inst jmp :e err-lab))
-                     ((plusp min)
-                      (inst cmp nargs (fixnumize min))
-                      (inst jmp :b err-lab)))))
-        (cond ((not min)
-               (if (zerop max)
-                   (inst test nargs nargs)
-                   (inst cmp nargs (fixnumize max)))
-               (inst jmp :ne err-lab))
-              (max
-               (check-min)
-               (inst cmp nargs (fixnumize max))
-               (inst jmp :a err-lab))
-              (t
-               (check-min)))))))
+      (cond ((not min)
+             (if (zerop max)
+                 (inst test nargs nargs)
+                 (inst cmp nargs (fixnumize max)))
+             (inst jmp :ne err-lab))
+            (max
+             (if (zerop min)
+                 (setf temp nargs)
+                 (inst lea temp (ea (fixnumize (- min)) nargs)))
+             (inst cmp temp (fixnumize (- max min)))
+             (inst jmp :a err-lab))
+            (t
+             (cond ((= min 1)
+                    (inst test nargs nargs)
+                    (inst jmp :e err-lab))
+                   ((plusp min)
+                    (inst cmp nargs (fixnumize min))
+                    (inst jmp :b err-lab))))))))
 
 ;; Signal an error about an untagged number.
 ;; These are pretty much boilerplate and could be generic except:
