@@ -924,6 +924,25 @@ constant shift greater than word length")))
     (inst sar result :cl)
     (inst and result (lognot fixnum-tag-mask))))
 
+(define-vop (fastrem-32)
+  (:translate fastrem-32)
+  (:policy :fast-safe)
+  (:args (dividend :scs (unsigned-reg) :target eax)
+         (c :scs (unsigned-reg))
+         (divisor :scs (unsigned-reg)))
+  (:arg-types unsigned-num unsigned-num unsigned-num)
+  (:results (remainder :scs (unsigned-reg)))
+  (:result-types unsigned-num)
+  (:temporary (:sc unsigned-reg :offset eax-offset
+               :from (:argument 0) :to (:result 0)) eax)
+  (:temporary (:sc unsigned-reg :offset edx-offset
+               :from (:argument 0) :to (:result 0)) edx)
+  (:generator 10
+    (move eax dividend)
+    (inst mul eax c) ; result to EDX:EAX (but we expressly drop all bits from EDX)
+    (inst mul eax divisor) ; new we want _only_ bits from rDX
+    (move remainder edx)))
+
 (in-package "SB-C")
 
 (defknown %lea (integer integer (member 1 2 4 8) (signed-byte 32))
@@ -1351,30 +1370,23 @@ constant shift greater than word length")))
 ;;;; 32-bit logical operations
 
 ;;; Only the lower 5 bits of the shift amount are significant.
-(define-vop (shift-towards-someplace)
-  (:policy :fast-safe)
-  (:args (num :scs (unsigned-reg) :target r)
-         (amount :scs (signed-reg) :target ecx))
-  (:arg-types unsigned-num tagged-num)
-  (:temporary (:sc signed-reg :offset ecx-offset :from (:argument 1)) ecx)
-  (:results (r :scs (unsigned-reg) :from (:argument 0)))
-  (:result-types unsigned-num))
-
-(define-vop (shift-towards-start shift-towards-someplace)
-  (:translate shift-towards-start)
-  (:note "SHIFT-TOWARDS-START")
-  (:generator 1
-    (move r num)
-    (move ecx amount)
-    (inst shr r :cl)))
-
-(define-vop (shift-towards-end shift-towards-someplace)
-  (:translate shift-towards-end)
-  (:note "SHIFT-TOWARDS-END")
-  (:generator 1
-    (move r num)
-    (move ecx amount)
-    (inst shl r :cl)))
+(macrolet ((define (translate operation)
+             `(define-vop ()
+                (:translate ,translate)
+                (:note ,(string translate))
+                (:policy :fast-safe)
+                (:args (num :scs (unsigned-reg) :target r)
+                       (amount :scs (signed-reg) :target ecx))
+                (:arg-types unsigned-num tagged-num)
+                (:temporary (:sc signed-reg :offset ecx-offset :from (:argument 1)) ecx)
+                (:results (r :scs (unsigned-reg) :from (:argument 0)))
+                (:result-types unsigned-num)
+                (:generator 1
+                 (move r num)
+                 (move ecx amount)
+                 (inst ,operation r :cl)))))
+  (define shift-towards-start shr)
+  (define shift-towards-end   shl))
 
 ;;;; Modular functions
 (defmacro define-mod-binop ((name prototype) function)
@@ -1472,7 +1484,7 @@ constant shift greater than word length")))
              fast-ash-left/unsigned=>unsigned))
 (deftransform ash-left-mod32 ((integer count)
                               ((unsigned-byte 32) (unsigned-byte 5)))
-  (when (sb-c::constant-lvar-p count)
+  (when (sb-c:constant-lvar-p count)
     (sb-c::give-up-ir1-transform))
   '(%primitive fast-ash-left-mod32/unsigned=>unsigned integer count))
 
@@ -1485,7 +1497,7 @@ constant shift greater than word length")))
              fast-ash-left/fixnum=>fixnum))
 (deftransform ash-left-modfx ((integer count)
                               (fixnum (unsigned-byte 5)))
-  (when (sb-c::constant-lvar-p count)
+  (when (sb-c:constant-lvar-p count)
     (sb-c::give-up-ir1-transform))
   '(%primitive fast-ash-left-modfx/fixnum=>fixnum integer count))
 
@@ -1595,13 +1607,17 @@ constant shift greater than word length")))
   (:translate sb-bignum:%bignum-set-length)
   (:policy :fast-safe))
 
+#-bignum-assertions ; %BIGNUM-ref is an inline function if compiling with assertions
 (define-full-reffer bignum-ref * bignum-digits-offset other-pointer-lowtag
   (unsigned-reg) unsigned-num sb-bignum:%bignum-ref)
-(define-full-reffer+offset bignum-ref-with-offset *
+#-bignum-assertions ; does not get called if compiling with assertions
+(define-full-reffer+addend bignum-ref-with-offset *
   bignum-digits-offset other-pointer-lowtag
   (unsigned-reg) unsigned-num sb-bignum:%bignum-ref-with-offset)
 (define-full-setter bignum-set * bignum-digits-offset other-pointer-lowtag
-  (unsigned-reg) unsigned-num sb-bignum:%bignum-set)
+  (unsigned-reg) unsigned-num
+  #+bignum-assertions sb-bignum:%%bignum-set
+  #-bignum-assertions sb-bignum:%bignum-set)
 
 (define-vop (digit-0-or-plus)
   (:translate sb-bignum:%digit-0-or-plusp)
@@ -1613,18 +1629,18 @@ constant shift greater than word length")))
     (inst test digit digit)))
 
 
-;;; For add and sub with carry the sc of carry argument is any-reg so
-;;; that it may be passed as a fixnum or word and thus may be 0, 1, or
-;;; 4. This is easy to deal with and may save a fixnum-word
-;;; conversion.
+;;; For add and sub with carry, the sc of carry argument is unsigned-reg
+;;; or any-reg so that it may be passed either as tagged or untagged.
+;;; This is easy to deal with and may save a fixnum-word conversion.
 (define-vop (add-w/carry)
   (:translate sb-bignum:%add-with-carry)
   (:policy :fast-safe)
   (:args (a :scs (unsigned-reg) :target result)
          (b :scs (unsigned-reg unsigned-stack) :to :eval)
-         (c :scs (any-reg) :target temp))
+         (c :scs (any-reg unsigned-reg control-stack) :target temp))
   (:arg-types unsigned-num unsigned-num positive-fixnum)
   (:temporary (:sc any-reg :from (:argument 2) :to :eval) temp)
+  (:temporary (:sc any-reg :from :eval :to :result :offset eax-offset :target carry) carry-temp)
   (:results (result :scs (unsigned-reg) :from (:argument 0))
             (carry :scs (unsigned-reg)))
   (:result-types unsigned-num positive-fixnum)
@@ -1633,8 +1649,9 @@ constant shift greater than word length")))
     (move temp c)
     (inst neg temp) ; Set the carry flag to 0 if c=0 else to 1
     (inst adc result b)
-    (inst mov carry 0)
-    (inst adc carry carry)))
+    (inst set :c carry-temp)
+    (inst and carry-temp 1)
+    (move carry carry-temp)))
 
 ;;; Note: the borrow is 1 for no borrow and 0 for a borrow, the opposite
 ;;; of the x86 convention.
@@ -1643,7 +1660,7 @@ constant shift greater than word length")))
   (:policy :fast-safe)
   (:args (a :scs (unsigned-reg) :to :eval :target result)
          (b :scs (unsigned-reg unsigned-stack) :to :result)
-         (c :scs (any-reg control-stack)))
+         (c :scs (any-reg unsigned-reg control-stack)))
   (:arg-types unsigned-num unsigned-num positive-fixnum)
   (:results (result :scs (unsigned-reg) :from :eval)
             (borrow :scs (unsigned-reg)))
