@@ -270,8 +270,6 @@
 
 (defgeneric effective-slot-definition-class (class &rest initargs))
 
-(defgeneric make-boundp-method-function (class slot-name))
-
 (defgeneric make-reader-method-function (class slot-name))
 
 (defgeneric make-writer-method-function (class slot-name))
@@ -300,8 +298,6 @@ evaluated (in contrast to DEFMETHOD's behavior).
 NOTE: This generic function is part of an SBCL-specific experimental
 protocol. Interface subject to change."))
 
-(defgeneric remove-boundp-method (class generic-function))
-
 (defgeneric remove-dependent (metaobject dependent))
 
 (defgeneric remove-direct-method (specializer method))
@@ -446,10 +442,6 @@ protocol. Interface subject to change."))
 
 ;;;; 5 arguments
 
-;;; FIXME: This is currently unused -- where should we call it? Or should we just
-;;; delete it.
-(defgeneric add-boundp-method (class generic-function slot-name slot-documentation source-location))
-
 (defgeneric add-reader-method (class generic-function slot-name slot-documentation source-location))
 
 (defgeneric add-writer-method (class generic-function slot-name slot-documentation source-location))
