@@ -1,4 +1,4 @@
-;;;; Bits and pieces of the wrapper machninery. This used to live in cache.lisp,
+;;;; Bits and pieces of the wrapper machinery. This used to live in cache.lisp,
 ;;;; but doesn't really logically belong there.
 
 ;;;; This software is part of the SBCL system. See the README file for
@@ -43,7 +43,7 @@
        (make-layout (hash-layout-name name)
                     (make-standard-classoid :name name)
                     :length length
-                    :flags +pcl-object-layout-flag+
+                    :flags (logior +pcl-object-layout-flag+ +strictly-boxed-flag+)
                     :bitmap bitmap
                     :invalid nil)))))
 
@@ -73,7 +73,8 @@
                       (bug "Got to T branch in ~S" 'make-wrapper))))))
         (make-layout (hash-layout-name name) classoid
                      :bitmap +layout-all-tagged+
-                     :invalid nil :length length :flags +pcl-object-layout-flag+)))
+                     :invalid nil :length length
+                     :flags (logior +pcl-object-layout-flag+ +strictly-boxed-flag+))))
     (t
      (let* ((found (find-classoid (slot-value class 'name)))
             (layout (classoid-wrapper found)))
@@ -111,7 +112,7 @@
 
 ;;; The portable code inherited from the ancient PCL sources used an extremely
 ;;; clever mechanism to store a mapping from new layouts ("wrappers") to old.
-;;; Old layouts point to new in a straighforward way - via the INVALID slot -
+;;; Old layouts point to new in a straightforward way - via the INVALID slot -
 ;;; but the invalidation logic wants backpointers from new to old as well.
 ;;; It arranges so that if the following state exists:
 ;;;   layout_A -> (:flush layout_B)
@@ -120,7 +121,7 @@
 ;;; the state of layout_A should be changed to (:flush layout_C)
 ;;; at the same time that the state of layout_B is changed.
 ;;;
-;;; In general, alterating layout_B might might have to perform more than one
+;;; In general, altering layout_B might might have to perform more than one
 ;;; update to older versions of the layout, so it's a one-to-many relation.
 ;;; The same "transitivity" could have been implemented at the time of flushing
 ;;; layout_A by chasing an extra pointer (the B -> C pointer) and recording that
@@ -128,12 +129,12 @@
 ;;; improves anything, as it merely performs work eagerly that could have
 ;;; been performed lazily. (PCL generally prefers lazy updates.)
 ;;;
-;;; Anyway the extreme cleverness (and premature optimization, imho), was almost,
+;;; Anyway the extreme cleverness (and premature optimization, IMHO), was almost,
 ;;; but not quite, clever enough - it leaked three conses per redefinition,
 ;;; which it pretty much had to in order to remain portable.
 ;;; We can at least replicate it in a way that doesn't leak conses,
 ;;; or at least theoretically doesn't. A likely scenario is that old caches may
-;;; never get adquately flushed to drop all references to obsolete layouts.
+;;; never get adequately flushed to drop all references to obsolete layouts.
 ;;; If you call a GF once and not again, it won't flush; thus causing a retained
 ;;; old layout no matter what else is done.
 ;;;
@@ -150,13 +151,13 @@
 ;;; The bug is that it was possible for an observer to see (:FLUSH new-thing)
 ;;; in an older layout when it should actually have seen (:OBSOLETE newer-thing).
 ;;; On the other hand, it's still possible for an observer to see the
-;;; wrong value in the INVALID slot itself. i.e. if it reads the slot
+;;; wrong value in the INVALID slot itself. I.e. if it reads the slot
 ;;; before %invalidate-wrapper performs the transitivity operation,
 ;;; hence the possibility that this optimization is not only premature,
 ;;; but also pretty much not correct.
 (defun %invalidate-wrapper (owrapper state nwrapper)
   (aver (member state '(:flush :obsolete)))
-  #+sb-thread (aver (sb-thread:holding-mutex-p sb-c::**world-lock**))
+  #+sb-thread (aver (sb-thread:holding-mutex-p sb-kernel::**world-lock**))
   (let ((classoid (wrapper-classoid nwrapper))
         (new-previous ())
         (new-state (cons state nwrapper)))
