@@ -25,11 +25,6 @@
 #include "monitor.h"
 #include "pseudo-atomic.h"
 
-void arch_init(void)
-{
-    return;
-}
-
 os_vm_address_t arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *context)
 {
     return (os_vm_address_t)code->si_addr;
@@ -37,9 +32,9 @@ os_vm_address_t arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *contex
 
 void arch_skip_instruction(os_context_t *context)
 {
-    u32 trap_instruction = *((u32 *)*os_context_pc_addr(context));
+    uint32_t trap_instruction = *(uint32_t *)OS_CONTEXT_PC(context);
     unsigned code = trap_instruction >> 5 & 0xFF;
-    *os_context_pc_addr(context) += 4;
+    OS_CONTEXT_PC(context) += 4;
     switch (code)
     {
     case trap_Error:
@@ -52,7 +47,7 @@ void arch_skip_instruction(os_context_t *context)
 
 unsigned char *arch_internal_error_arguments(os_context_t *context)
 {
-    return (unsigned char *)*os_context_pc_addr(context);
+    return (unsigned char *)OS_CONTEXT_PC(context);
 }
 
 boolean arch_pseudo_atomic_atomic(os_context_t *context)
@@ -68,7 +63,7 @@ boolean arch_pseudo_atomic_atomic(os_context_t *context)
      * to arch_pseudo_atomic_atomic, but this seems clearer.
      * --NS 2007-05-15 */
 #ifdef LISP_FEATURE_GENCGC
-    return get_pseudo_atomic_atomic(arch_os_get_current_thread());
+    return get_pseudo_atomic_atomic(get_sb_vm_thread());
 #else
     return (!foreign_function_call_active)
         && (NIL != SymbolValue(PSEUDO_ATOMIC_ATOMIC,0));
@@ -77,12 +72,12 @@ boolean arch_pseudo_atomic_atomic(os_context_t *context)
 
 void arch_set_pseudo_atomic_interrupted(os_context_t *context)
 {
-    set_pseudo_atomic_interrupted(arch_os_get_current_thread());
+    set_pseudo_atomic_interrupted(get_sb_vm_thread());
 }
 
 void arch_clear_pseudo_atomic_interrupted(os_context_t *context)
 {
-    clear_pseudo_atomic_interrupted(arch_os_get_current_thread());
+    clear_pseudo_atomic_interrupted(get_sb_vm_thread());
 }
 
 unsigned int arch_install_breakpoint(void *pc)
@@ -111,7 +106,7 @@ arch_handle_breakpoint(os_context_t *context)
 void
 arch_handle_fun_end_breakpoint(os_context_t *context)
 {
-    *os_context_pc_addr(context) = (int) handle_fun_end_breakpoint(context);
+    OS_CONTEXT_PC(context) = (uword_t) handle_fun_end_breakpoint(context);
 }
 
 void
@@ -124,33 +119,49 @@ arch_handle_single_step_trap(os_context_t *context, int trap)
 static void
 sigtrap_handler(int signal, siginfo_t *siginfo, os_context_t *context)
 {
-    u32 trap_instruction = *((u32 *)*os_context_pc_addr(context));
+    uint32_t trap_instruction = *(uint32_t *)OS_CONTEXT_PC(context);
     unsigned code = trap_instruction >> 5 & 0xFF;
     if ((trap_instruction >> 21) != 0x6A1) {
-        lose("Unrecognized trap instruction %08lx in sigtrap_handler() (PC: %p)",
-             trap_instruction, *os_context_pc_addr(context));
+        lose("Unrecognized trap instruction %08x in sigtrap_handler() (PC: %p)",
+             trap_instruction, (void*)OS_CONTEXT_PC(context));
     }
 
     handle_trap(context, code);
 }
 
+void sigill_handler(int signal, siginfo_t *siginfo, os_context_t *context);
+
+#ifndef LISP_FEATURE_DARWIN
+void
+sigill_handler(int signal, siginfo_t *siginfo, os_context_t *context) {
+    fake_foreign_function_call(context);
+    lose("Unhandled SIGILL at %p.", (void*)OS_CONTEXT_PC(context));
+}
+#endif
+
 void arch_install_interrupt_handlers()
 {
-    undoably_install_low_level_interrupt_handler(SIGTRAP, sigtrap_handler);
+    ll_install_handler(SIGTRAP, sigtrap_handler);
+    ll_install_handler(SIGILL, sigill_handler);
 }
 
 
-#ifdef LISP_FEATURE_LINKAGE_TABLE
-
 /* Linkage tables
  *
  * Linkage entry size is 16, because we need 2 instructions and an 8 byte address.
  */
 
-#define LINKAGE_TEMP_REG        reg_NFP
+#define LINKAGE_TEMP_REG reg_NL9
 
-void arch_write_linkage_table_jmp(char *reloc_addr, void *target_addr)
+void arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
+  THREAD_JIT(0);
+  char *reloc_addr = (char*)ALIEN_LINKAGE_TABLE_SPACE_START + index * ALIEN_LINKAGE_TABLE_ENTRY_SIZE;
+
+  if (datap) {
+    *(unsigned long *)reloc_addr = (unsigned long)target_addr;
+    goto DONE;
+  }
   /*
     ldr reg,=address
     br  reg
@@ -170,15 +181,10 @@ void arch_write_linkage_table_jmp(char *reloc_addr, void *target_addr)
   *inst_ptr++ = inst;
 
   // address
-  *(unsigned long *)inst_ptr++ = target_addr;
+  *(unsigned long *)inst_ptr++ = (unsigned long)target_addr;
 
   os_flush_icache((os_vm_address_t) reloc_addr, (char*) inst_ptr - reloc_addr);
-}
 
-void
-arch_write_linkage_table_ref(void * reloc_addr, void *target_addr)
-{
-    *(unsigned long *)reloc_addr = (unsigned long)target_addr;
+ DONE:
+  THREAD_JIT(1);
 }
-
-#endif
