@@ -19,6 +19,7 @@
 #include "genesis/vector.h"
 #include "genesis/layout.h"
 #include "genesis/hash-table.h"
+#include "genesis/list-node.h"
 #include "code.h"
 #include "immobile-space.h"
 #include "queue.h"
@@ -151,6 +152,10 @@ static inline int pointer_survived_gc_yet(lispobj pointer)
     return (fullcgcmarks[mark_index / N_WORD_BITS] >> (mark_index % N_WORD_BITS)) & 1;
 }
 
+int fullcgc_lispobj_livep(lispobj pointer) {
+    return pointer_survived_gc_yet(pointer);
+}
+
 void dump_marked_objects() {
     fprintf(stderr, "Marked objects:\n");
     page_index_t first = 0;
@@ -247,8 +252,8 @@ static void trace_using_layout(lispobj layout, lispobj* where, int nslots)
     gc_mark_obj(layout);
 #endif
     if (lockfree_list_node_layout_p(LAYOUT(layout))) { // allow untagged 'next'
-        struct instance* node = (struct instance*)where;
-        lispobj next = node->slots[INSTANCE_DATA_START];
+        struct list_node* node = (void*)where;
+        lispobj next = node->_node_next;
         // ignore if 0
         if (fixnump(next) && next) __mark_obj(next|INSTANCE_POINTER_LOWTAG);
     }
@@ -589,6 +594,7 @@ void execute_full_sweep_phase()
     local_smash_weak_pointers();
     gc_dispose_private_pages();
     cull_weak_hash_tables(alivep_funs);
+    scan_finalizers();
 
     memset(words_zeroed, 0, sizeof words_zeroed);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
