@@ -26,27 +26,15 @@
 #include "code.h"
 #include "gc-internal.h"
 #include "gc-private.h"
-#include <stdarg.h>
+#include "genesis/gc-tables.h"
 #include "thread.h"              /* genesis/primitive-objects.h needs this */
 #include <errno.h>
 #include <stdlib.h>
 #include <inttypes.h>
 #include <setjmp.h>
 
-/* FSHOW and odxprint provide debugging output for low-level information
- * (signal handling, exceptions, safepoints) which is hard to debug by
- * other means.
- *
- * If enabled at all, environment variables control whether calls of the
- * form odxprint(name, ...) are enabled at run-time, e.g. using
- * SBCL_DYNDEBUG="fshow fshow_signal safepoints".
- *
- * In the case of FSHOW and FSHOW_SIGNAL, old-style code from runtime.h
- * can also be used to enable or disable these more aggressively.
- */
-
 struct dyndebug_config dyndebug_config = {
-    QSHOW == 2, QSHOW_SIGNALS == 2
+    QSHOW == 2,
 };
 
 void
@@ -67,7 +55,6 @@ dyndebug_init()
     int *ptrs[DYNDEBUG_NFLAGS];
 
     dyndebug_init1(fshow,          "FSHOW");
-    dyndebug_init1(fshow_signal,   "FSHOW_SIGNAL");
     dyndebug_init1(gencgc_verbose, "GENCGC_VERBOSE");
     dyndebug_init1(safepoints,     "SAFEPOINTS");
     dyndebug_init1(seh,            "SEH");
@@ -130,81 +117,6 @@ dyndebug_init()
 #undef DYNDEBUG_NFLAGS
 }
 
-/* Temporarily, odxprint merely performs the equivalent of a traditional
- * FSHOW call, i.e. it merely formats to stderr.  Ultimately, it should
- * be restored to its full win32 branch functionality, where output to a
- * file or to the debugger can be selected at runtime. */
-
-void vodxprint_fun(const char *, va_list);
-
-void
-odxprint_fun(const char *fmt, ...)
-{
-    va_list args;
-    va_start(args, fmt);
-    vodxprint_fun(fmt, args);
-    va_end(args);
-}
-
-void
-vodxprint_fun(const char *fmt, va_list args)
-{
-#ifdef LISP_FEATURE_WIN32
-    DWORD lastError = GetLastError();
-#endif
-    int original_errno = errno;
-
-    char buf[1024];
-    int n = 0;
-
-#ifdef LISP_FEATURE_SB_THREAD
-    snprintf(buf, sizeof(buf), "["THREAD_ID_LABEL"] ", THREAD_ID_VALUE);
-    n = strlen(buf);
-#endif
-
-    vsnprintf(buf + n, sizeof(buf) - n - 1, fmt, args);
-    /* buf is now zero-terminated (even in case of overflow).
-     * Our caller took care of the newline (if any) through `fmt'. */
-
-    /* A sufficiently POSIXy implementation of stdio will provide
-     * per-FILE locking, as defined in the spec for flockfile.  At least
-     * glibc complies with this.  Hence we do not need to perform
-     * locking ourselves here.  (Should it turn out, of course, that
-     * other libraries opt for speed rather than safety, we need to
-     * revisit this decision.) */
-    fputs(buf, stderr);
-
-#ifdef LISP_FEATURE_WIN32
-    /* stdio's stderr is line-bufferred, i.e. \n ought to flush it.
-     * Unfortunately, MinGW does not behave the way I would expect it
-     * to.  Let's be safe: */
-    fflush(stderr);
-#endif
-
-#ifdef LISP_FEATURE_WIN32
-    SetLastError(lastError);
-#endif
-    errno = original_errno;
-}
-
-/* Translate the rather awkward syntax
- *   FSHOW((stderr, "xyz"))
- * into the new and cleaner
- *   odxprint("xyz").
- * If we were willing to clean up all existing call sites, we could remove
- * this wrapper function.  (This is a function, because I don't know how to
- * strip the extra parens in a macro.) */
-void
-fshow_fun(void __attribute__((__unused__)) *ignored,
-          const char *fmt,
-          ...)
-{
-    va_list args;
-    va_start(args, fmt);
-    vodxprint_fun(fmt, args);
-    va_end(args);
-}
-
 #include "monitor.h"
 #include "vars.h"
 #include "os.h"
@@ -394,6 +306,22 @@ static void brief_list(lispobj obj)
     }
 }
 
+void print_list_car_ptrs(lispobj obj, FILE* f)
+{
+    char sep = '(';
+    int len = 0;
+    if (obj == NIL) { fprintf(f, "NIL"); return; }
+    do {
+        if (++len > 20) { fprintf(f, "...)"); return; }
+        fprintf(f, "%c%p", sep, (void*)CONS(obj)->car);
+        obj = CONS(obj)->cdr;
+        sep = ' ';
+    } while (listp(obj) && obj != NIL);
+    if (obj != NIL) fprintf(f, " . %p", (void*)obj);
+    putc(')', f);
+}
+
+
 static void print_list(lispobj obj)
 {
     if (obj == NIL) {
@@ -421,7 +349,7 @@ char * simple_base_stringize(struct vector * string)
 
 static void brief_struct(lispobj obj)
 {
-    struct instance *instance = (struct instance *)native_pointer(obj);
+    struct instance *instance = INSTANCE(obj);
     extern struct vector * instance_classoid_name(lispobj*);
     struct vector * classoid_name;
     classoid_name = instance_classoid_name((lispobj*)instance);
@@ -464,7 +392,7 @@ static boolean tagged_slot_p(struct layout *layout, int slot_index)
 
 static void print_struct(lispobj obj)
 {
-    struct instance *instance = (struct instance *)native_pointer(obj);
+    struct instance *instance = INSTANCE(obj);
     short int i;
     char buffer[16];
     lispobj layout = instance_layout(native_pointer(obj));
@@ -515,7 +443,6 @@ void show_lstring(struct vector * string, int quotes, FILE *s)
 
 static void brief_fun_or_otherptr(lispobj obj)
 {
-    extern void safely_show_lstring(struct vector*, int, FILE*);
     lispobj *ptr, header;
     int type;
     struct symbol *symbol;
@@ -526,9 +453,10 @@ static void brief_fun_or_otherptr(lispobj obj)
     switch (type) {
         case SYMBOL_WIDETAG:
             symbol = (struct symbol *)ptr;
-            if (symbol->package == NIL)
+            lispobj package = symbol_package(symbol);
+            if (package == NIL)
                 printf("#:");
-            show_lstring(VECTOR(symbol->name), 0, stdout);
+            show_lstring(symbol_name(symbol), 0, stdout);
             break;
 
         case SIMPLE_BASE_STRING_WIDETAG:
@@ -547,7 +475,7 @@ static void brief_fun_or_otherptr(lispobj obj)
                 if (lowtag_of(name) == OTHER_POINTER_LOWTAG
                     && widetag_of(native_pointer(name)) == SYMBOL_WIDETAG) {
                   printf(" for ");
-                  struct vector* str = symbol_name(native_pointer(name));
+                  struct vector* str = symbol_name(SYMBOL(name));
                   safely_show_lstring(str, 0, stdout);
                 }
             }
@@ -559,10 +487,14 @@ static void print_slots(char **slots, int count, lispobj *ptr)
 {
     while (count-- > 0) {
         if (*slots) {
-            // kludge for half-lispword sized slot
-            print_obj(*slots,
-                      (N_WORD_BYTES == 8 && !strcmp(*slots, "boxed_size: "))
-                      ? *ptr & 0xFFFFFFFF : *ptr);
+            // kludge for encoded slots
+            lispobj word = *ptr;
+            char* slot_name = *slots;
+            if (N_WORD_BYTES == 8 && !strcmp(slot_name, "boxed_size: ")) word = word & 0xFFFFFFFF;
+#ifdef LISP_FEATURE_COMPACT_SYMBOL
+            else if (!strcmp(slot_name, "name: ")) word = decode_symbol_name(word);
+#endif
+            print_obj(slot_name, word);
             slots++;
         } else {
             print_obj("???: ", *ptr);
@@ -571,23 +503,9 @@ static void print_slots(char **slots, int count, lispobj *ptr)
     }
 }
 
-lispobj symbol_function(lispobj* symbol)
+lispobj symbol_function(struct symbol* symbol)
 {
-    lispobj info = ((struct symbol*)symbol)->info;
-    if (listp(info))
-        info = CONS(info)->cdr;
-    if (lowtag_of(info) == OTHER_POINTER_LOWTAG) {
-        struct vector* v = VECTOR(info);
-        int len = vector_len(v);
-        if (len != 0) {
-            lispobj elt = v->data[0];  // Just like INFO-VECTOR-FDEFN
-            if (fixnump(elt) && (fixnum_value(elt) & 07777) >= 07701) {
-                lispobj fdefn = v->data[len-1];
-                if (lowtag_of(fdefn) == OTHER_POINTER_LOWTAG)
-                    return FDEFN(fdefn)->fun;
-            }
-        }
-    }
+    if (symbol->fdefn) return FDEFN(symbol->fdefn)->fun;
     return NIL;
 }
 
@@ -617,6 +535,7 @@ static void print_fun_or_otherptr(lispobj obj)
 
     switch (type) {
     case BIGNUM_WIDETAG:
+        count &= 0x7fffff;
         ptr += count;
         NEWLINE_OR_RETURN;
         printf("0x");
@@ -642,15 +561,20 @@ static void print_fun_or_otherptr(lispobj obj)
         // Only 1 byte of a symbol header conveys its size.
         // The other bytes may be freely used by the backend.
         print_slots(symbol_slots, count & 0xFF, ptr);
-        if (symbol_function(ptr-1) != NIL)
-            print_obj("fun: ", symbol_function(ptr-1));
+        struct symbol* sym = (void*)(ptr - 1);
+        if (symbol_function(sym) != NIL) print_obj("fun: ", symbol_function(sym));
 #ifdef LISP_FEATURE_SB_THREAD
-        int tlsindex = tls_index_of((struct symbol*)(ptr-1));
+        int tlsindex = tls_index_of(sym);
         struct thread*th = get_sb_vm_thread();
         if (th != 0 && tlsindex != 0) {
             lispobj v = *(lispobj*)(tlsindex + (char*)th);
             print_obj("tlsval: ", v);
         }
+#endif
+#ifdef LISP_FEATURE_COMPACT_SYMBOL
+        // print_obj doesn't understand raw words, so make it a fixnum
+        int pkgid = symbol_package_id(sym) << N_FIXNUM_TAG_BITS;
+        print_obj("package_id: ", pkgid);
 #endif
         break;
 
@@ -800,7 +724,7 @@ static void print_fun_or_otherptr(lispobj obj)
 
     case FDEFN_WIDETAG:
         print_slots(fdefn_slots, 2, ptr);
-        print_obj("entry: ", fdefn_callee_lispobj((struct fdefn*)(ptr-1)));
+        print_obj("entry: ", decode_fdefn_rawfun((struct fdefn*)(ptr-1)));
         break;
 
     // Make some vectors printable from C, for when all hell breaks lose
@@ -820,9 +744,8 @@ static void print_fun_or_otherptr(lispobj obj)
         break;
     default:
         NEWLINE_OR_RETURN;
-        if (type >= SIMPLE_ARRAY_UNSIGNED_BYTE_2_WIDETAG &&
-            type <= SIMPLE_BIT_VECTOR_WIDETAG) // ASSUMPTION: widetag ordering
-            printf("length = %ld", vector_len(VECTOR(obj)));
+        if (specialized_vector_widetag_p(type))
+            printf("length = %"OBJ_FMTd, vector_len(VECTOR(obj)));
         else
             printf("Unknown header object?");
         break;
@@ -922,22 +845,22 @@ void brief_print(lispobj obj)
 // hence all the forwarding checks.
 
 #include "forwarding-ptr.h"
-#include "genesis/classoid.h"
-struct vector * symbol_name(lispobj * sym)
+struct vector * symbol_name(struct symbol* sym)
 {
-  if (forwarding_pointer_p(sym))
-    sym = native_pointer(forwarding_pointer_value(sym));
-  if (lowtag_of(((struct symbol*)sym)->name) != OTHER_POINTER_LOWTAG)
-      return NULL;
-  return VECTOR(follow_maybe_fp(((struct symbol*)sym)->name));
+  if (forwarding_pointer_p((lispobj*)sym))
+    sym = (void*)native_pointer(forwarding_pointer_value((lispobj*)sym));
+  lispobj name = sym->name;
+  if (lowtag_of(name) != OTHER_POINTER_LOWTAG) return NULL;
+  lispobj string = decode_symbol_name(name);
+  return VECTOR(follow_maybe_fp(string));
 }
 struct vector * classoid_name(lispobj * classoid)
 {
   if (forwarding_pointer_p(classoid))
       classoid = native_pointer(forwarding_pointer_value(classoid));
-  lispobj sym = ((struct classoid*)classoid)->name;
-  return lowtag_of(sym) != OTHER_POINTER_LOWTAG ? NULL
-    : symbol_name(native_pointer(sym));
+  // Classoids are named by symbols even though a CLASS name is arbitrary (theoretically)
+  lispobj sym = classoid[CLASSOID_NAME_WORDINDEX];
+  return lowtag_of(sym) != OTHER_POINTER_LOWTAG ? NULL : symbol_name(SYMBOL(sym));
 }
 struct vector * layout_classoid_name(lispobj * layout)
 {
